from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from diffusers import AutoencoderKL, FlowMatchEulerDiscreteScheduler, SD3Transformer2DModel
from diffusers.image_processor import PipelineImageInput
from diffusers.models.autoencoders.vae import DiagonalGaussianDistribution
from diffusers.models.downsampling import Downsample2D
from diffusers.models.upsampling import Upsample2D
from diffusers.pipelines.stable_diffusion_3.pipeline_output import StableDiffusion3PipelineOutput
from diffusers.utils.torch_utils import randn_tensor
from torch.nn import MSELoss
from transformers import PreTrainedModel
from transformers.modeling_outputs import ModelOutput

from .....utils import logging
from ....seed_omni.projector import build_feature_projector
from .configuration_ultra_edit import UltraEditConfig


logger = logging.get_logger(__name__)


def retrieve_latents(
    encoder_output: torch.Tensor, generator: Optional[torch.Generator] = None, sample_mode: str = "sample"
):
    if hasattr(encoder_output, "latent_dist") and sample_mode == "sample":
        return encoder_output.latent_dist.sample(generator)
    elif hasattr(encoder_output, "latent_dist") and sample_mode == "argmax":
        return encoder_output.latent_dist.mode()
    elif hasattr(encoder_output, "latents"):
        return encoder_output.latents
    else:
        raise AttributeError("Could not access latents of provided encoder_output")


@dataclass
class UltraEditOutput(ModelOutput):
    loss: Optional[torch.FloatTensor] = None
    logits: torch.FloatTensor = None


class ConvGenerationHead(nn.Module):
    def __init__(
        self,
        feature_size: int,
        hidden_size: int,
        output_size: int,
        kernel_size1: int = 14,
        kernel_size2: int = 2,
        restore_size: int = 512,
    ) -> None:
        super().__init__()
        self.proj = build_feature_projector(feature_size, hidden_size * 4)
        self.kernel_size1 = kernel_size1
        self.kernel_size2 = kernel_size2
        self.restore_size = restore_size
        self.hidden_size = hidden_size
        self.block_out_channels = [8, 16, 32, 64]
        self.upsampler = Upsample2D(hidden_size, use_conv=True, out_channels=self.block_out_channels[0])
        blocks = []
        for index in range(1, len(self.block_out_channels)):
            in_channel = self.block_out_channels[index - 1]
            out_channel = self.block_out_channels[index]
            blocks.append(nn.GroupNorm(num_groups=in_channel // 4, num_channels=in_channel, eps=1e-6))
            blocks.append(nn.SiLU())
            blocks.append(Downsample2D(in_channel, use_conv=True, out_channels=out_channel))
        blocks += [
            nn.GroupNorm(
                num_groups=self.block_out_channels[-1] // 4, num_channels=self.block_out_channels[-1], eps=1e-6
            ),
            nn.SiLU(),
            nn.Conv2d(self.block_out_channels[-1], output_size * 2, kernel_size=3, padding=1),
        ]
        self.downsampler = nn.Sequential(*blocks)

    def forward(self, x, image_grid_thw):
        x = self.proj(x)
        x = x.view(-1, self.hidden_size)
        x = x.transpose(1, 0).reshape(x.shape[0], -1, 1, 1)
        x = self.upsampler(x, output_size=(self.kernel_size1, self.kernel_size1))
        cu_index, restored_x = 0, []
        for grid_thw in image_grid_thw:
            h_factor = grid_thw[1]
            w_factor = grid_thw[2]
            token_num = h_factor * w_factor
            latent_x = x[cu_index : cu_index + token_num].reshape(
                h_factor // 2, w_factor // 2, x.shape[1], x.shape[2] * 2, x.shape[3] * 2
            )
            cu_index += token_num
            latent_x = latent_x.permute(2, 0, 3, 1, 4).reshape(
                1, x.shape[1], h_factor * self.kernel_size1, w_factor * self.kernel_size1
            )
            pad_w = (self.restore_size - w_factor * self.kernel_size1) // 2
            pad_h = (self.restore_size - h_factor * self.kernel_size1) // 2
            latent_x = F.pad(latent_x, (pad_w, pad_w, pad_h, pad_h), mode="constant", value=0)
            restored_x.append(latent_x)

        restored_x = torch.cat(restored_x, dim=0)
        latent = self.downsampler(restored_x)
        posterior = DiagonalGaussianDistribution(latent)
        kl_loss = posterior.kl()
        latent = posterior.mode()
        return latent, kl_loss.mean()


class UltraEdit(PreTrainedModel):
    config_class = UltraEditConfig
    _supports_flash_attn_2 = True
    supports_gradient_checkpointing = False

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.ConvTranspose2d, nn.Conv2d)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            # init.xavier_uniform(module.weight.data)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=0.02)
            # init.xavier_uniform(module.weight.data)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.GroupNorm):
            if module.weight is not None:
                module.weight.data.fill_(1.0)
            if module.bias is not None:
                module.bias.data.zero_()

    def __init__(self, config: UltraEditConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.config = config
        if config.transformer_config is not None:
            self.transformer = SD3Transformer2DModel.from_config(config.transformer_config)
            self.vae = AutoencoderKL.from_config(config.vae_config)
            self.noise_scheduler = FlowMatchEulerDiscreteScheduler.from_config(config.scheduler_config)
        else:
            self.transformer, self.vae, self.noise_scheduler = None, None, None
            logger.info_rank0("Make sure you are converting checkpoints instead of training/inferencing.")

        if config.output_size is not None:
            self.gen_head = build_feature_projector(config.output_size, config.output_size)
            self.proj_head = ConvGenerationHead(
                config.output_size,
                config.proj_hidden_dim,
                config.condition_dim,
                kernel_size1=config.patch_size1,
                kernel_size2=config.patch_size2,
                restore_size=512,
            )
        else:
            self.gen_head = nn.Identity()
            self.proj_head = nn.Identity()

        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1) if self.vae is not None else 8
        self.default_sample_size = self.transformer.config.sample_size if self.transformer is not None else 128
        self.dummy_text_embeds = nn.Parameter(torch.zeros(1, 154, 4096), requires_grad=False)
        self.dummy_pooled_text_embeds = nn.Parameter(torch.zeros(1, 2048), requires_grad=False)
        # self.image_processor = VaeImageProcessor(vae_scale_factor=self.vae_scale_factor)

    def prepare_latents(self, batch_size, num_channels_latents, height, width, dtype, device, generator, latents=None):
        if latents is not None:
            return latents.to(device=device, dtype=dtype)
        shape = (
            batch_size,
            num_channels_latents,
            int(height) // self.vae_scale_factor,
            int(width) // self.vae_scale_factor,
        )
        if isinstance(generator, list) and len(generator) != batch_size:
            raise ValueError(
                f"You have passed a list of generators of length {len(generator)}, but requested an effective batch"
                f" size of {batch_size}. Make sure the batch size matches the length of the generators."
            )
        latents = randn_tensor(shape, generator=generator, device=device, dtype=dtype)
        return latents

    def _get_t5_prompt_embeds(
        self,
        prompt: Union[str, List[str]] = None,
        text_input_ids: Optional[torch.Tensor] = None,
        num_images_per_prompt: int = 1,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None,
        text_encoder=None,
        tokenizer=None,
    ):
        if text_input_ids is None:
            prompt = [prompt] if isinstance(prompt, str) else prompt
            text_inputs = tokenizer(
                prompt,
                padding="max_length",
                max_length=77,
                truncation=True,
                add_special_tokens=True,
                return_tensors="pt",
            )
            text_input_ids = text_inputs.input_ids
        batch_size = text_input_ids.shape[0]
        prompt_embeds = text_encoder(text_input_ids.to(device))[0]
        prompt_embeds = prompt_embeds.to(dtype=dtype, device=device)

        _, seq_len, _ = prompt_embeds.shape

        # duplicate text embeddings and attention mask for each generation per prompt, using mps friendly method
        prompt_embeds = prompt_embeds.repeat(1, num_images_per_prompt, 1)
        prompt_embeds = prompt_embeds.view(batch_size * num_images_per_prompt, seq_len, -1)

        return prompt_embeds

    def _get_clip_prompt_embeds(
        self,
        prompt: Union[str, List[str]] = None,
        text_input_ids: Optional[torch.Tensor] = None,
        num_images_per_prompt: int = 1,
        device: Optional[torch.device] = None,
        clip_skip: Optional[int] = None,
        text_encoder=None,
        tokenizer=None,
    ):
        if text_input_ids is None:
            prompt = [prompt] if isinstance(prompt, str) else prompt

            text_inputs = tokenizer(
                prompt,
                padding="max_length",
                max_length=tokenizer.model_max_length,
                truncation=True,
                return_tensors="pt",
            )

            text_input_ids = text_inputs.input_ids
        batch_size = text_input_ids.shape[0]
        prompt_embeds = text_encoder(text_input_ids.to(device), output_hidden_states=True)
        pooled_prompt_embeds = prompt_embeds[0]

        if clip_skip is None:
            prompt_embeds = prompt_embeds.hidden_states[-2]
        else:
            prompt_embeds = prompt_embeds.hidden_states[-(clip_skip + 2)]

        prompt_embeds = prompt_embeds.to(dtype=text_encoder.dtype, device=device)

        _, seq_len, _ = prompt_embeds.shape
        # duplicate text embeddings for each generation per prompt, using mps friendly method
        prompt_embeds = prompt_embeds.repeat(1, num_images_per_prompt, 1)
        prompt_embeds = prompt_embeds.view(batch_size * num_images_per_prompt, seq_len, -1)

        pooled_prompt_embeds = pooled_prompt_embeds.repeat(1, num_images_per_prompt, 1)
        pooled_prompt_embeds = pooled_prompt_embeds.view(batch_size * num_images_per_prompt, -1)

        return prompt_embeds, pooled_prompt_embeds

    def encode_prompt(
        self,
        prompt: Union[str, List[str]] = None,
        input_ids: Optional[List[torch.Tensor]] = [None, None, None],
        device: Optional[torch.device] = None,
        num_images_per_prompt: int = 1,
        do_classifier_free_guidance: bool = True,
        prompt_embeds: Optional[torch.FloatTensor] = None,
        negative_prompt_embeds: Optional[torch.FloatTensor] = None,
        pooled_prompt_embeds: Optional[torch.FloatTensor] = None,
        negative_pooled_prompt_embeds: Optional[torch.FloatTensor] = None,
        clip_skip: Optional[int] = None,
        text_encoders=[None, None, None],
        tokenizers=[None, None, None],
    ):
        prompt = [prompt] if isinstance(prompt, str) else prompt
        if prompt_embeds is None:
            prompt_embed, pooled_prompt_embed = self._get_clip_prompt_embeds(
                prompt=prompt,
                text_input_ids=input_ids[0],
                device=device,
                num_images_per_prompt=num_images_per_prompt,
                clip_skip=clip_skip,
                text_encoder=text_encoders[0],
                tokenizer=tokenizers[0],
            )
            prompt_2_embed, pooled_prompt_2_embed = self._get_clip_prompt_embeds(
                prompt=prompt,
                text_input_ids=input_ids[1],
                device=device,
                num_images_per_prompt=num_images_per_prompt,
                clip_skip=clip_skip,
                text_encoder=text_encoders[1],
                tokenizer=tokenizers[1],
            )
            clip_prompt_embeds = torch.cat([prompt_embed, prompt_2_embed], dim=-1)

            t5_prompt_embed = self._get_t5_prompt_embeds(
                prompt=prompt,
                text_input_ids=input_ids[2],
                num_images_per_prompt=num_images_per_prompt,
                device=device,
                tokenizer=tokenizers[2],
                text_encoder=text_encoders[2],
            )

            clip_prompt_embeds = torch.nn.functional.pad(
                clip_prompt_embeds, (0, t5_prompt_embed.shape[-1] - clip_prompt_embeds.shape[-1])
            )

            prompt_embeds = torch.cat([clip_prompt_embeds, t5_prompt_embed], dim=-2)
            pooled_prompt_embeds = torch.cat([pooled_prompt_embed, pooled_prompt_2_embed], dim=-1)

        if do_classifier_free_guidance and negative_prompt_embeds is None:
            negative_prompt_embed, negative_pooled_prompt_embed = self._get_clip_prompt_embeds(
                prompt=[""],
                device=device,
                num_images_per_prompt=num_images_per_prompt,
                clip_skip=None,
                text_encoder=text_encoders[0],
                tokenizer=tokenizers[0],
            )
            negative_prompt_2_embed, negative_pooled_prompt_2_embed = self._get_clip_prompt_embeds(
                prompt=[""],
                device=device,
                num_images_per_prompt=num_images_per_prompt,
                clip_skip=None,
                text_encoder=text_encoders[1],
                tokenizer=tokenizers[1],
            )
            negative_clip_prompt_embeds = torch.cat([negative_prompt_embed, negative_prompt_2_embed], dim=-1)

            t5_negative_prompt_embed = self._get_t5_prompt_embeds(
                prompt=[""],
                num_images_per_prompt=num_images_per_prompt,
                device=device,
                tokenizer=tokenizers[2],
                text_encoder=text_encoders[2],
            )

            negative_clip_prompt_embeds = torch.nn.functional.pad(
                negative_clip_prompt_embeds,
                (0, t5_negative_prompt_embed.shape[-1] - negative_clip_prompt_embeds.shape[-1]),
            )

            negative_prompt_embeds = torch.cat([negative_clip_prompt_embeds, t5_negative_prompt_embed], dim=-2)
            negative_pooled_prompt_embeds = torch.cat(
                [negative_pooled_prompt_embed, negative_pooled_prompt_2_embed], dim=-1
            )

        return prompt_embeds, negative_prompt_embeds, pooled_prompt_embeds, negative_pooled_prompt_embeds

    def prepare_image_latents(
        self, image, batch_size, num_images_per_prompt, dtype, device, do_classifier_free_guidance, generator=None
    ):
        image = image.to(device=device, dtype=dtype)
        batch_size = batch_size * num_images_per_prompt

        if image.shape[1] == self.vae.config.latent_channels:
            image_latents = image
        else:
            image_latents = retrieve_latents(self.vae.encode(image), sample_mode="argmax")
            # ? normalize image latents
            # image_latents = (image_latents - self.vae.config.shift_factor) * self.vae.config.scaling_factor

        image_latents = torch.cat([image_latents], dim=0)

        if do_classifier_free_guidance:
            uncond_image_latents = torch.zeros_like(image_latents)
            image_latents = torch.cat([image_latents, image_latents, uncond_image_latents], dim=0)

        return image_latents

    def set_projector_trainable_only(self):
        self.requires_grad_(False)
        self.gen_head.requires_grad_(True)
        self.proj_head.requires_grad_(True)

    @property
    def guidance_scale(self):
        return self._guidance_scale

    @property
    def image_guidance_scale(self):
        return self._image_guidance_scale

    @property
    def clip_skip(self):
        return self._clip_skip

    @property
    def do_classifier_free_guidance(self):
        return self.guidance_scale > 1.0 and self.image_guidance_scale >= 1.0

    @property
    def joint_attention_kwargs(self):
        return self._joint_attention_kwargs

    @property
    def num_timesteps(self):
        return self._num_timesteps

    @property
    def interrupt(self):
        return self._interrupt

    def get_sigmas(self, timesteps, n_dim, dtype):
        sigmas = self.noise_scheduler.sigmas.to(device=self.device, dtype=dtype)
        schedule_timesteps = self.noise_scheduler.timesteps.to(self.device)
        step_indices = [(schedule_timesteps == t).nonzero().item() for t in timesteps]

        sigma = sigmas[step_indices].flatten()
        while len(sigma.shape) < n_dim:
            sigma = sigma.unsqueeze(-1)
        return sigma

    def lm_head(self, hidden_states, labels=None, features=None, image_grid_thw=None, mask=None, **kwargs):
        pred_feature = self.gen_head(hidden_states)

        loss = None
        if labels is not None:
            loss_func = MSELoss()
            mse_loss = loss_func(pred_feature, labels)

            original_image_embeds, kl_loss = self.proj_head(pred_feature, image_grid_thw)
            latents = self.vae.encode(features).latent_dist.mode()
            latents = latents * self.vae.config.scaling_factor
            # Sample noise that we'll add to the latents
            noise = torch.randn_like(latents)
            bsz = latents.shape[0]
            # Sample a random timestep for each image
            # for weighting schemes where we sample timesteps non-uniformly
            # See 3.1 in the SD3 paper ($rf/lognorm(0.00,1.00)$).
            u = torch.normal(mean=0.0, std=0.1, size=(bsz,))
            u = torch.nn.functional.sigmoid(u)

            indices = (u * self.noise_scheduler.config.num_train_timesteps).long()
            timesteps = self.noise_scheduler.timesteps[indices].to(self.device)

            # Add noise to the latents according to the noise magnitude at each timestep
            # (this is the forward diffusion process)
            sigmas = self.get_sigmas(timesteps, n_dim=latents.ndim, dtype=latents.dtype)
            noisy_model_input = sigmas * noise + (1.0 - sigmas) * latents
            concatenated_noisy_latents = torch.cat([noisy_model_input, original_image_embeds], dim=1)

            mask_embeds = self.vae.encode(mask).latent_dist.mode()
            concatenated_noisy_latents = torch.cat([concatenated_noisy_latents, mask_embeds], dim=1)

            prompt_embeds = self.dummy_text_embeds.repeat(bsz, 1, 1)
            pooled_prompt_embeds = self.dummy_pooled_text_embeds.repeat(bsz, 1)

            model_pred = self.transformer(
                hidden_states=concatenated_noisy_latents,
                timestep=timesteps,
                encoder_hidden_states=prompt_embeds,
                pooled_projections=pooled_prompt_embeds,
                return_dict=False,
            )[0]

            model_pred = model_pred * (-sigmas) + noisy_model_input
            weighting = torch.ones_like(sigmas)

            target = latents
            # Get the target for loss depending on the prediction type
            diff_loss = torch.mean(
                (weighting.float() * (model_pred.float() - target.float()) ** 2).reshape(target.shape[0], -1), 1
            )
            loss = mse_loss + torch.mean(diff_loss) + kl_loss

        return UltraEditOutput(
            loss=loss,
            logits=pred_feature,
        )

    def forward(
        self,
        origin_pixel_values,
        edited_pixel_values,
        mask_pixel_values,
        prompt_embeds,
        pooled_prompt_embeds,
        **kwargs,
    ):
        latents = self.vae.encode(edited_pixel_values).latent_dist.sample()
        latents = latents * self.vae.config.scaling_factor
        # Sample noise that we'll add to the latents
        noise = torch.randn_like(latents)
        bsz = latents.shape[0]
        # Sample a random timestep for each image
        # for weighting schemes where we sample timesteps non-uniformly
        # See 3.1 in the SD3 paper ($rf/lognorm(0.00,1.00)$).
        u = torch.normal(mean=0.0, std=0.1, size=(bsz,))
        u = torch.nn.functional.sigmoid(u)

        indices = (u * self.noise_scheduler.config.num_train_timesteps).long()
        timesteps = self.noise_scheduler.timesteps[indices].to(self.device)

        # Add noise to the latents according to the noise magnitude at each timestep
        # (this is the forward diffusion process)
        sigmas = self.get_sigmas(timesteps, n_dim=latents.ndim, dtype=latents.dtype)
        noisy_model_input = sigmas * noise + (1.0 - sigmas) * latents

        # Get the additional image embedding for conditioning.
        # Instead of getting a diagonal Gaussian here, we simply take the mode.
        original_image_embeds = self.vae.encode(origin_pixel_values).latent_dist.mode()
        concatenated_noisy_latents = torch.cat([noisy_model_input, original_image_embeds], dim=1)

        mask_embeds = self.vae.encode(mask_pixel_values).latent_dist.mode()
        concatenated_noisy_latents = torch.cat([concatenated_noisy_latents, mask_embeds], dim=1)

        model_pred = self.transformer(
            hidden_states=concatenated_noisy_latents,
            timestep=timesteps,
            encoder_hidden_states=prompt_embeds,
            pooled_projections=pooled_prompt_embeds,
            return_dict=False,
        )[0]

        model_pred = model_pred * (-sigmas) + noisy_model_input
        weighting = torch.ones_like(sigmas)

        target = latents
        # Get the target for loss depending on the prediction type
        loss = torch.mean(
            (weighting.float() * (model_pred.float() - target.float()) ** 2).reshape(target.shape[0], -1), 1
        )

        return torch.mean(loss)

    @torch.no_grad()
    def generate(
        self,
        image: PipelineImageInput = None,
        num_inference_steps: int = 50,
        timesteps: List[int] = None,
        guidance_scale: float = 7.5,
        image_guidance_scale: float = 1.5,
        num_images_per_prompt: Optional[int] = 1,
        generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
        latents: Optional[torch.FloatTensor] = None,
        prompt_embeds: Optional[torch.FloatTensor] = None,
        negative_prompt_embeds: Optional[torch.FloatTensor] = None,
        pooled_prompt_embeds: Optional[torch.FloatTensor] = None,
        negative_pooled_prompt_embeds: Optional[torch.FloatTensor] = None,
        output_type: Optional[str] = "pil",
        joint_attention_kwargs: Optional[Dict[str, Any]] = None,
        clip_skip: Optional[int] = None,
        **kwargs,
    ):
        self._guidance_scale = guidance_scale
        self._image_guidance_scale = image_guidance_scale
        self._clip_skip = clip_skip
        self._joint_attention_kwargs = joint_attention_kwargs
        self._interrupt = False

        batch_size = prompt_embeds.shape[0]

        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0 and image_guidance_scale >= 1.0

        if do_classifier_free_guidance:
            # The extra concat similar to how it's done in SD InstructPix2Pix.
            prompt_embeds = torch.cat([prompt_embeds, negative_prompt_embeds, negative_prompt_embeds], dim=0)
            pooled_prompt_embeds = torch.cat(
                [pooled_prompt_embeds, negative_pooled_prompt_embeds, negative_pooled_prompt_embeds], dim=0
            )
        # 1. Prepare timesteps
        self.noise_scheduler.set_timesteps(num_inference_steps, device=self.device)
        timesteps = self.noise_scheduler.timesteps
        # 2. Prepare Image latents
        pad_height = (512 - image.shape[2]) // 2
        pad_width = (512 - image.shape[3]) // 2
        mask_img = torch.ones_like(image)
        padding = (pad_width, pad_width, pad_height, pad_height)
        image = F.pad(image, padding, mode="constant", value=0)
        mask_img = F.pad(mask_img, padding, mode="constant", value=0)
        image_latents = self.prepare_image_latents(
            image,
            batch_size,
            num_images_per_prompt,
            prompt_embeds.dtype,
            self.device,
            do_classifier_free_guidance,
        )

        height, width = image_latents.shape[-2:]
        height = height * self.vae_scale_factor
        width = width * self.vae_scale_factor

        # 3. Prepare latent variables
        num_channels_latents = self.vae.config.latent_channels
        latents = self.prepare_latents(
            batch_size * num_images_per_prompt,
            num_channels_latents,
            height,
            width,
            prompt_embeds.dtype,
            self.device,
            generator,
            latents,
        )

        num_channels_image = image_latents.shape[1]
        if mask_img is not None:
            mask_img = self.image_processor.preprocess(mask_img)
            mask_image_latents = self.prepare_image_latents(
                mask_img,
                batch_size,
                num_images_per_prompt,
                prompt_embeds.dtype,
                self.device,
                self.do_classifier_free_guidance,
            )
            num_channels_image += mask_image_latents.shape[1]

        # 4. Check that shapes of latents and image match the UNet channels
        if num_channels_latents + num_channels_image != self.transformer.config.in_channels:
            raise ValueError(
                f"Incorrect configuration settings! The config of `pipeline.unet`: {self.transformer.config} expects"
                f" {self.transformer.config.in_channels} but received `num_channels_latents`: {num_channels_latents} +"
                f" `num_channels_image`: {num_channels_image} "
                f" = {num_channels_latents + num_channels_image}. Please verify the config of"
                " `pipeline.unet` or your `image` input."
            )

        # 5. Denoising loop
        for i, t in enumerate(timesteps):
            # Expand the latents if we are doing classifier free guidance.
            # The latents are expanded 3 times because for pix2pix the guidance
            # is applied for both the text and the input image.
            latent_model_input = torch.cat([latents] * 3) if do_classifier_free_guidance else latents
            timestep = t.expand(latent_model_input.shape[0])

            # concat latents, image_latents in the channel dimension
            scaled_latent_model_input = torch.cat([latent_model_input, image_latents], dim=1)
            if mask_img is not None:
                scaled_latent_model_input = torch.cat([scaled_latent_model_input, mask_image_latents], dim=1)
            # predict the noise residual
            noise_pred = self.transformer(
                scaled_latent_model_input,
                timestep=timestep,
                encoder_hidden_states=prompt_embeds,
                pooled_projections=pooled_prompt_embeds,
                joint_attention_kwargs=self.joint_attention_kwargs,
                return_dict=False,
            )[0]

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_text, noise_pred_image, noise_pred_uncond = noise_pred.chunk(3)
                noise_pred = (
                    noise_pred_uncond
                    + guidance_scale * (noise_pred_text - noise_pred_image)
                    + image_guidance_scale * (noise_pred_image - noise_pred_uncond)
                )

            # compute the previous noisy sample x_t -> x_t-1
            latents_dtype = latents.dtype
            latents = self.noise_scheduler.step(noise_pred, t, latents, return_dict=False)[0]
            if latents.dtype != latents_dtype:
                if torch.backends.mps.is_available():
                    # some platforms (eg. apple mps) misbehave due to a pytorch bug: https://github.com/pytorch/pytorch/pull/99272
                    latents = latents.to(latents_dtype)
        if not output_type == "latent":
            latents = latents / self.vae.config.scaling_factor
            latent_height = pad_height // self.vae_scale_factor
            latent_width = pad_width // self.vae_scale_factor
            latents = latents[:, :, latent_height:-latent_height, latent_width:-latent_width]
            image = self.vae.decode(latents, return_dict=False)[0]
            image = self.image_processor.postprocess(image, output_type=output_type)
            return StableDiffusion3PipelineOutput(images=image)
        else:
            return StableDiffusion3PipelineOutput(images=latents)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ...loader import MODEL_CONFIG_REGISTRY, MODELING_REGISTRY


@MODEL_CONFIG_REGISTRY.register("qwen3_moe")
def register_qwen3_moe_config():
    from .configuration_qwen3_moe import Qwen3MoeConfig

    return Qwen3MoeConfig


@MODELING_REGISTRY.register("qwen3_moe")
def register_qwen3_moe_modeling(architecture: str):
    from .modeling_qwen3_moe import (
        Qwen3MoeForCausalLM,
        Qwen3MoeForQuestionAnswering,
        Qwen3MoeModel,
    )

    if "ForCausalLM" in architecture:
        return Qwen3MoeForCausalLM
    elif "ForQuestionAnswering" in architecture:
        return Qwen3MoeForQuestionAnswering
    elif "Model" in architecture:
        return Qwen3MoeModel
    else:
        return Qwen3MoeForCausalLM



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import types
from functools import partial
from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
from torch.distributed._tensor import Shard
from torch.distributed.fsdp import CPUOffload, FullyShardedDataParallel, MixedPrecision, ShardingStrategy
from torch.distributed.fsdp._common_utils import _get_module_fsdp_state_if_fully_sharded_module
from torch.distributed.fsdp._runtime_utils import _lazy_init
from torch.distributed.fsdp.wrap import lambda_auto_wrap_policy
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.checkpoint import noop_context_fn

from ..models import load_model_weights, rank0_load_and_broadcast_weights
from ..utils import logging
from ..utils.device import IS_NPU_AVAILABLE, get_device_id, get_device_type
from ..utils.import_utils import is_torch_version_greater_than
from .checkpoint import CheckpointFunction
from .fsdp import (
    clip_grad_norm_,
    init_fsdp_fn,
    parallel_init_fsdp_fn,
    parallel_load_safetensors,
    register_checkpoint_extension,
)
from .parallel_state import get_parallel_state
from .utils import get_module_from_path, set_module_from_path


if is_torch_version_greater_than("2.4"):
    from torch.distributed._composable.fsdp import MixedPrecisionPolicy, fully_shard
    from torch.distributed.tensor.parallel import parallelize_module


logger = logging.get_logger(__name__)


def _reset_hf_initialized_flag(module: nn.Module) -> None:
    if hasattr(module, "_is_hf_initialized"):
        module._is_hf_initialized = False
    for child in module.children():
        _reset_hf_initialized_flag(child)


def verbose_fsdp_grouping(model, prefix="", depth=0):
    indent = "    " * depth

    for name, child in model.named_children():
        if isinstance(child, FullyShardedDataParallel):
            module_names = [m_name for m_name, _ in child.named_modules()][1:]  # [1:] 排除自身
            strategy = child.sharding_strategy
            logger.debug_rank0(f"{indent}├── [FSDP Group] {prefix}{name}")
            logger.debug_rank0(
                f"{indent}│   ├── Sharding Strategy: {strategy}, Mixed Precision: {child.mixed_precision}"
            )
            logger.debug_rank0(f"{indent}│   └── Contains Modules: {module_names}")

            verbose_fsdp_grouping(child, prefix=f"{prefix}{name}.", depth=depth + 1)
        else:
            verbose_fsdp_grouping(child, prefix=f"{prefix}{name}.", depth=depth)


def parallelize_model_fsdp1(
    model: "nn.Module",
    weights_path: Optional[str] = None,
    enable_full_shard: bool = True,
    enable_mixed_precision: bool = True,
    basic_modules: Optional[List[str]] = None,
    fsdp_no_shard_states=None,
    fsdp_no_shard_states_fqn=None,
    ep_param_suffix=None,
    fqn2spec_info=None,
    **kwargs,
) -> "nn.Module":
    """
    Applies EP (when enabled) + FSDP1 parallel strategy to the model.
    """
    parallel_state = get_parallel_state()

    if parallel_state.ep_enabled:
        parallel_plan = model.get_parallel_plan()
        fqn2spec_info = parallel_plan.apply(model, parallel_state.ep_fsdp_device_mesh)

        fsdp_no_shard_states_fqn_to_module = parallel_plan.get_fsdp_no_shard_info(model)
        fsdp_no_shard_states = list(fsdp_no_shard_states_fqn_to_module.values())
        fsdp_no_shard_states_fqn = list(fsdp_no_shard_states_fqn_to_module.keys())
        logger.info_rank0(f"Apply expert parallel to the model successfully.\nEP modules: {fsdp_no_shard_states_fqn}.")
    else:
        fqn2spec_info = None
        fsdp_no_shard_states = None
        fsdp_no_shard_states_fqn = None

    wrap_policy = partial(lambda_auto_wrap_policy, lambda_fn=lambda module: module.__class__.__name__ in basic_modules)

    # set fsdp/hsdp sharding strategy
    if parallel_state.fsdp_mesh.ndim > 1 and parallel_state.fsdp_mesh.size() > 1:
        strategy = ShardingStrategy.HYBRID_SHARD
    else:
        strategy = ShardingStrategy.FULL_SHARD

    fsdp_kwargs = {
        "auto_wrap_policy": wrap_policy,
        "ignored_states": fsdp_no_shard_states,
        "device_id": get_device_id(),
        "sharding_strategy": strategy if enable_full_shard else ShardingStrategy.NO_SHARD,
        "use_orig_params": True,
    }

    fsdp_kwargs["device_mesh"] = parallel_state.fsdp_mesh

    fsdp_kwargs.update(kwargs.pop("fsdp_kwargs", {}))

    if enable_mixed_precision:
        logger.info_rank0("Enable mixed precision training.")
        mixed_precision = MixedPrecision(
            param_dtype=torch.bfloat16,
            reduce_dtype=torch.float32,
            buffer_dtype=torch.float32,
        )
        if hasattr(model, "get_ignore_modules_in_mixed_precision"):
            mixed_precision._module_classes_to_ignore += model.get_ignore_modules_in_mixed_precision()

        fsdp_kwargs["mixed_precision"] = mixed_precision

    if kwargs.get("init_device") == "cpu":
        logger.info_rank0("Enable rank0-only initialization.")
        fsdp_kwargs["sync_module_states"] = True
        if parallel_state.global_rank != 0:
            fsdp_kwargs["param_init_fn"] = init_fsdp_fn(model, device=get_device_type())
    elif kwargs.get("init_device") == "meta":
        logger.info_rank0("Enable meta initialization.")
        if weights_path is None:
            logger.info_rank0("weights_path is None during meta initialization.")

        shard_states = kwargs.get("shard_states", {})
        if not shard_states and weights_path:
            shard_states = parallel_load_safetensors(weights_path)

        fsdp_kwargs["param_init_fn"] = parallel_init_fsdp_fn(
            model,
            shard_states.copy(),
            ignore_states=fsdp_no_shard_states,
            strict=kwargs.pop("strict", False),
        )

    if kwargs.pop("enable_fsdp_offload", False):
        logger.info_rank0("Enable offloading for parameters & gradients & optimizer states.")
        fsdp_kwargs["cpu_offload"] = CPUOffload(offload_params=True)

    if kwargs.pop("enable_forward_prefetch", False):
        fsdp_kwargs["forward_prefetch"] = True
    else:
        fsdp_kwargs["forward_prefetch"] = False
        fsdp_kwargs["backward_prefetch"] = None

    # FULLY_SHARD first
    model = FullyShardedDataParallel(model, **fsdp_kwargs)

    if fsdp_no_shard_states is not None:
        # apply NO_SHARD the ignored_states, but wrap into DDP
        if parallel_state.ep_fsdp_mesh["ep_fsdp"].size() == 1:
            moe_sharding_strategy = ShardingStrategy.NO_SHARD
            ep_fsdp_device_mesh = parallel_state.fsdp_mesh
        else:
            moe_sharding_strategy = ShardingStrategy.FULL_SHARD
            ep_fsdp_device_mesh = parallel_state.ep_fsdp_mesh["ep_fsdp"]

        logger.info_rank0(f"Apply {moe_sharding_strategy} states on '{fsdp_no_shard_states_fqn}'.")
        fsdp_kwargs.pop("ignored_states", None)
        fsdp_kwargs.pop("auto_wrap_policy", None)
        fsdp_kwargs["sharding_strategy"] = moe_sharding_strategy
        fsdp_kwargs["device_mesh"] = ep_fsdp_device_mesh
        logger.info_rank0(f"{ep_fsdp_device_mesh=}")
        for fqn in fsdp_no_shard_states_fqn:
            no_shard_module = get_module_from_path(model, fqn)
            if kwargs.get("init_device") == "meta":
                key_prefix = fqn + "."
                ep_shard_states = {
                    k[len(key_prefix) :]: v for k, v in shard_states.items() if k.startswith(key_prefix)
                }
                fsdp_kwargs["param_init_fn"] = parallel_init_fsdp_fn(
                    no_shard_module,
                    ep_shard_states,
                    strict=kwargs.pop("strict", False),
                )
            fsdp_module = FullyShardedDataParallel(no_shard_module, **fsdp_kwargs)
            fsdp_state = _get_module_fsdp_state_if_fully_sharded_module(fsdp_module)
            fsdp_state._gradient_postdivide_factor *= parallel_state.ep_size
            set_module_from_path(model, fqn, fsdp_module)

    _lazy_init(model, model)

    # Apply fsdp extension to FSDP model
    save_hook_mesh = parallel_state.ep_fsdp_device_mesh if parallel_state.ep_enabled else None
    logger.info_rank0("Register Checkpoints Extension hook to the model")
    register_checkpoint_extension(
        fsdp_model=model,
        save_hook_mesh=save_hook_mesh,
        fqn2spec_info=fqn2spec_info,
    )

    if parallel_state.ep_enabled:
        model.clip_grad_norm_ = types.MethodType(clip_grad_norm_, model)

    verbose_fsdp_grouping(model)

    # FSDP1 path keeps its own clipper already above

    return model


def parallelize_model_fsdp2(
    model: "nn.Module",
    weights_path: Optional[str] = None,
    enable_mixed_precision: bool = True,
    basic_modules: Optional[List[str]] = None,
    **kwargs,
) -> "nn.Module":
    """
    Applies EP (when enabled) + FSDP2 parallel strategy to the model.

    Flow:
    1. Apply EP: Expert tensors [128,H,I] -> [32,H,I] local tensors per EP rank
    2. Apply FSDP2 to expert modules: Shard expert tensors along dim-1 (hidden dim)
    3. Apply FSDP2 to regular modules: Standard dim-0 sharding
    4. Result: Expert params [32,H/fsdp_size,I], regular params use standard FSDP2
    """
    parallel_state = get_parallel_state()

    # Step 0: Get target classes to shard later
    target_classes = set((getattr(model, "_no_split_modules", []) or []) + (basic_modules or []))
    # Make a list of tuples that contains layer's name and module
    decoder_blocks: List[Tuple[str, nn.Module]] = [
        (fqn, mod) for fqn, mod in model.named_modules() if mod.__class__.__name__ in target_classes
    ]
    logger.info_rank0(f"target classes to shard: {target_classes}")

    # Step 1: Apply expert parallelism (slice expert tensors [128,H,I] -> [16,H,I])
    if parallel_state.ep_enabled:
        parallel_plan = model.get_parallel_plan()
        assert parallel_plan is not None, (
            "Expert parallelism needs parallel plan defined in the model! Please see veomni/models/transformers/qwen3_moe/parallel_plan.py for example."
        )
        ep_fqn2spec_info = parallel_plan.apply(model, parallel_state.ep_fsdp_device_mesh)
        # Attach spec mapping for checkpoint load-time reconstruction
        setattr(model, "_fqn2spec_info", ep_fqn2spec_info)
        # ep_mesh does not really exist in EP parameters' device mesh.
        # EP parameters are loaded as local tensors to be later sharded by fully_shard
        ep_mesh = parallel_state.ep_fsdp_device_mesh["ep"]
        # experts_map is a dict {experts_fqn: experts_mod}
        # For example, Qwen3MoE keys: model.layers.N.mlp.experts
        experts_map = parallel_plan.get_fsdp_no_shard_info(model)

        logger.info_rank0(f"Applied EP: expert tensors sliced along expert dimension (EP mesh: {ep_mesh})")
        logger.info_rank0(f"Experts Map: {experts_map}")
    else:
        ep_fqn2spec_info = None
        ep_mesh = None
        experts_map = None

    # Extract experts module from the layer if any, then pair them
    layer_pairs = []
    for layer_fqn, layer_mod in decoder_blocks:
        if experts_map is not None:
            # extract experts module from the layer
            experts_mod = next(
                (exp_mod for exp_fqn, exp_mod in experts_map.items() if exp_fqn.startswith(layer_fqn + ".")),
                None,
            )
            layer_pairs.append((layer_fqn, layer_mod, experts_mod))
        else:
            # No experts module found in this layer
            # this is often the case for models like deepseek in which some decoder layers are dense instead of MoE
            layer_pairs.append((layer_fqn, layer_mod, None))

    logger.info_rank0(f"layer pairs: {layer_pairs}")

    # Step 2: Update fsdp2 kwargs
    fsdp_kwargs = {"mesh": parallel_state.fsdp_mesh}
    # mp_policy kwargs
    if enable_mixed_precision:
        mp_policy = MixedPrecisionPolicy(
            param_dtype=torch.bfloat16,
            reduce_dtype=torch.float32,
        )
        fsdp_kwargs["mp_policy"] = mp_policy

    if hasattr(model, "get_ignore_modules_in_mixed_precision"):
        modules_to_ignore_in_mixed_precision = model.get_ignore_modules_in_mixed_precision()
    else:
        modules_to_ignore_in_mixed_precision = None

    if modules_to_ignore_in_mixed_precision:
        assert isinstance(modules_to_ignore_in_mixed_precision, tuple), (
            "modules_to_ignore_in_mixed_precision needs to be a tuple!"
        )
        mp_ignored_classes = modules_to_ignore_in_mixed_precision
        fsdp_kwargs_without_mp = dict(fsdp_kwargs)
        fsdp_kwargs_without_mp.pop("mp_policy", None)
        # for high-precision modules, we do not reshard them after forward to avoid all-gather them in backward
        # these modules will stay in GPU memory so please ensure high-precision modules do not contain too many parameters
        fsdp_kwargs_without_mp["reshard_after_forward"] = False
    else:
        mp_ignored_classes = None
        fsdp_kwargs_without_mp = fsdp_kwargs

    # prepare ep_fsdp2 kwargs
    if parallel_state.ep_enabled:
        # Use the ep_fsdp dimension as DP mesh for experts (shards orthogonal to EP)
        ep_fsdp_mesh = parallel_state.ep_fsdp_device_mesh["ep_fsdp"]
        expert_fsdp_kwargs = dict(fsdp_kwargs)
        expert_fsdp_kwargs["mesh"] = ep_fsdp_mesh

        # Prefer dim-1 sharding for expert weights when composing with EP shard on dim-0
        def _experts_shard_placement_fn(param):
            return Shard(1)

        expert_fsdp_kwargs["shard_placement_fn"] = _experts_shard_placement_fn

    # Here we have a basic assumption for the decoder layer hierarchy
    # Decoder Layer
    # | -- layers that are sharded by fully_shard(decode_layer) (e.g., Attention)
    # | -- experts layer (apply fully_shard separately in order to shard across EP groups on the same EP rank instead of sharding globally)
    # | -- layers (declared in model.modules_to_ignore_in_mixed_precision) that need to apply fully_shard separately due to different mp policy as the decoder layer
    #      (e.g., some models requires MoE TopK gate layer to have parameters in higher FP32 precision in forward).
    for layer_fqn, layer_mod, experts_mod in layer_pairs:
        # register all the FSDPModule inside this decoder layer for the convenience of manual prefetching configuration
        layer_mod._fsdp_modules = []
        # ep enabled and this layer contains the expert module
        if parallel_state.ep_enabled and experts_mod is not None:
            # shard expert
            fully_shard(experts_mod, **expert_fsdp_kwargs)
            # average EP grads across EP ranks
            # NOTE: in torch 2.8 and later we should use
            # experts_mod.set_gradient_divide_factor(parallel_state.ep_size)
            # but now for torch 2.7 compatability we still use this legacy API
            # Note that NPU does not support PreSumMul so we skip this call
            # TODO(https://github.com/ByteDance-Seed/VeOmni/issues/241):
            # NPU is missing PreSumMul ReduceOp for `set_gradient_divide_factor` API
            # As a result, we will handle the grad dividing inisde the grad norm clipping
            # Need to remove this condition after the issue is resolved.
            if not IS_NPU_AVAILABLE:
                gradient_divide_factor = parallel_state.ep_gradient_divide_factor
                logger.info(f"setting grad divide factor for ep module to {gradient_divide_factor}")
                experts_mod.set_gradient_divide_factor(gradient_divide_factor)
            layer_mod._fsdp_modules.append(experts_mod)
        # shard module that needs to ignore mixed precision control
        if mp_ignored_classes:
            for sub_mod in layer_mod.modules():
                if isinstance(sub_mod, mp_ignored_classes) and sub_mod is not layer_mod:
                    fully_shard(sub_mod, **fsdp_kwargs_without_mp)
                    layer_mod._fsdp_modules.append(sub_mod)

        # shard everything else in the decoder layer
        fully_shard(layer_mod, **fsdp_kwargs)
        layer_mod._fsdp_modules.append(layer_mod)
        logger.info_rank0(f"{layer_fqn=}, {layer_mod._fsdp_modules=}")
    # shard root model
    fully_shard(model, **fsdp_kwargs)

    # configure manual prefetching when needed
    need_manual_prefetch = parallel_state.ep_enabled or mp_ignored_classes is not None
    if need_manual_prefetch:
        blocks = [pair[1] for pair in layer_pairs]
        next_blocks = blocks[1:] + [None]
        for current_block, next_block in zip(blocks, next_blocks):
            if next_block is not None:
                prefetch_modules = next_block._fsdp_modules
                # prefetch in order of attn, gate, experts
                current_block.set_modules_to_forward_prefetch(list(reversed(prefetch_modules)))

        # configure backward prefetch
        rev_blocks = list(reversed(blocks))
        prev_blocks = rev_blocks[1:] + [None]
        for current_block, prev_block in zip(rev_blocks, prev_blocks):
            if prev_block is not None:
                prefetch_modules = prev_block._fsdp_modules
                current_block.set_modules_to_backward_prefetch(list(reversed(prefetch_modules)))

    # Handle meta initialization for FSDP2 (fallback if pre-load not done)
    assert kwargs.get("init_device") == "meta", "Please use init_device: meta for FSDP2"

    if weights_path is None:
        model.to_empty(device=get_device_type())
        _reset_hf_initialized_flag(model)
        model.init_weights()
    else:
        from torch.distributed.tensor import distribute_tensor

        logger.info_rank0("starting to load model weights...")
        if kwargs.get("broadcast_model_weights_from_rank0"):
            logger.info_rank0("Loading model weights from disk on rank0 then broadcasting to other ranks...")
            rank0_load_and_broadcast_weights(model, weights_path, get_device_type(), dtensor_factory=distribute_tensor)
        else:
            logger.info_rank0("Every rank would read weights from disk and expect this to be slow!")
            load_model_weights(model, weights_path, get_device_type(), dtensor_factory=distribute_tensor)

    # Register grad norm clipping method for FSDP2
    from .fsdp2 import clip_grad_norm as clip_grad_norm_fn

    model.clip_grad_norm_ = types.MethodType(clip_grad_norm_fn, model)

    return model


def build_parallelize_model(
    model: "nn.Module",
    weights_path: Optional[str] = None,
    sharding_plan: Optional[Dict[str, Any]] = None,
    enable_full_shard: bool = True,
    enable_mixed_precision: bool = True,
    enable_gradient_checkpointing: bool = True,
    basic_modules: Optional[List[str]] = None,
    **kwargs,
) -> "nn.Module":
    """
    Applies parallel strategies to the model.
    """
    parallel_state = get_parallel_state()
    fsdp_no_shard_states = None

    if not parallel_state.fsdp_enabled:
        if kwargs.get("init_device") not in ["cuda", "npu"]:
            raise ValueError("Only FSDP training supports `init_device=cpu` or `init_device=meta`.")
        if kwargs.pop("enable_fsdp_offload", False):
            raise ValueError("Only FSDP training supports `enable_fsdp_offload`.")

    if enable_mixed_precision:  # upcast to float32 before feed it to optimizer
        model = model.float()

    if enable_gradient_checkpointing and hasattr(model, "gradient_checkpointing_enable"):
        logger.info_rank0("Enable gradient checkpointing.")
        use_reentrant = kwargs.pop("enable_reentrant", False)
        if use_reentrant:
            torch.utils.checkpoint.CheckpointFunction = CheckpointFunction

        model.gradient_checkpointing_enable(
            gradient_checkpointing_kwargs={
                "use_reentrant": use_reentrant,
                "context_fn": kwargs.pop("recompute_context_fn", noop_context_fn),
            },
        )

    if parallel_state.tp_enabled:
        logger.info_rank0("Apply tensor parallel to the model.")
        model = parallelize_module(
            model,
            device_mesh=parallel_state.tp_mesh,
        )

    if parallel_state.fsdp_enabled:
        logger.info_rank0(f"Apply data parallel to the model: {parallel_state.dp_mode}.")
        if parallel_state.dp_mode == "fsdp2":
            model = parallelize_model_fsdp2(
                model=model,
                weights_path=weights_path,
                enable_full_shard=enable_full_shard,
                enable_mixed_precision=enable_mixed_precision,
                basic_modules=basic_modules,
                **kwargs,
            )
        elif parallel_state.dp_mode == "fsdp1":
            model = parallelize_model_fsdp1(
                model=model,
                weights_path=weights_path,
                enable_full_shard=enable_full_shard,
                enable_mixed_precision=enable_mixed_precision,
                basic_modules=basic_modules,
                fsdp_no_shard_states=fsdp_no_shard_states,
                **kwargs,
            )
        else:
            ddp_kwargs = {"device_ids": [parallel_state.local_rank]}
            if enable_mixed_precision:
                logger.info_rank0("Enable mixed precision training.")
                mixed_precision = MixedPrecision(
                    param_dtype=torch.bfloat16,
                    reduce_dtype=torch.float32,
                    buffer_dtype=torch.bfloat16,
                )
                ddp_kwargs["mixed_precision"] = mixed_precision

            model = DDP(model, **ddp_kwargs)

    return model



# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os
from typing import Callable, Dict, Optional, Tuple

import torch
import torch.nn as nn
from einops import rearrange
from transformers.modeling_utils import PreTrainedModel

from veomni.distributed.parallel_state import get_parallel_state
from veomni.distributed.sequence_parallel import (
    gather_heads_scatter_seq,
    gather_outputs,
    gather_seq_scatter_heads,
    get_ulysses_sequence_parallel_world_size,
    slice_input_tensor_scale_grad,
)

from ....utils import logging
from ....utils.import_utils import is_liger_kernel_available
from .config_wan import WanConfig


if is_liger_kernel_available():
    from liger_kernel.transformers.rms_norm import LigerRMSNorm

from transformers.modeling_utils import ALL_ATTENTION_FUNCTIONS

from veomni.distributed.sequence_parallel.async_ulysses_dit import (
    async_ulysses_output_projection,
    async_ulysses_qkv_projection,
)


logger = logging.get_logger(__name__)

try:
    import flash_attn_interface

    FLASH_ATTN_3_AVAILABLE = True
except ModuleNotFoundError:
    FLASH_ATTN_3_AVAILABLE = False


try:
    from sageattention import sageattn

    SAGE_ATTN_AVAILABLE = True
except ModuleNotFoundError:
    SAGE_ATTN_AVAILABLE = False


def stochastic_round_tensor(x: torch.Tensor) -> torch.Tensor:
    """
    Perform stochastic rounding on a tensor
    Args:
        x (torch.Tensor): Input tensor
    Returns:
        torch.Tensor: Stochastically rounded integer tensor
    """
    floor_x = torch.floor(x)
    frac = x - floor_x
    rand_vals = torch.rand_like(x)
    round_up = rand_vals < frac
    result = floor_x + round_up.to(x.dtype)
    return result


def symmetric_quantize(x, dtype=torch.float8_e4m3fn):
    """
    Dynamic symmetric quantization that supports block-wise quantization under multi-head attention mechanism
    Args:
        x: Input tensor [batch_size, seq_len, head_count, head_dim]
        dtype: Target quantization type, defaults to torch.float8_e4m3fn
    Returns:
        x_quantized: Quantized tensor [batch_size, seq_len, head_count, head_dim]
        scales: Scaling factors for each head [batch_size, head_count]
    """
    batch_size, seq_len, head_count, head_dim = x.shape
    x = x.to(torch.float32)
    max_vals = x.abs().amax(dim=(1, 3), keepdim=True)  # [batch, 1, head, 1]
    finfo = torch.finfo(dtype)
    eps = 1e-12  # Smaller epsilon for better stability
    scales = (max_vals + eps) / finfo.max  # Ensure non-zero denominator
    scales = scales.clamp(min=eps)
    x_scaled = x / scales
    is_round = True
    if is_round:
        x_rounded = stochastic_round_tensor(x_scaled)
        x_clamped = x_rounded.clamp(min=finfo.min, max=finfo.max)
    else:
        x_clamped = x_scaled.clamp(min=finfo.min, max=finfo.max)
    x_quantized = x_clamped.to(dtype)
    scales = scales.squeeze((1, 3)).to(torch.float32)
    return x_quantized, scales


def modulate(x: torch.Tensor, shift: torch.Tensor, scale: torch.Tensor):
    return x * (1 + scale) + shift


def rearrange_qkv(q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, rerange_type: str, head_dim: int):
    q = rearrange(q, rerange_type, d=head_dim)
    k = rearrange(k, rerange_type, d=head_dim)
    v = rearrange(v, rerange_type, d=head_dim)
    return q, k, v


def gather_seq_scatter_heads_qkv(q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, seq_dim: int, head_dim: int):
    q = gather_seq_scatter_heads(q, seq_dim, head_dim)
    k = gather_seq_scatter_heads(k, seq_dim, head_dim)
    v = gather_seq_scatter_heads(v, seq_dim, head_dim)
    return q, k, v


def eager_attention_forward(
    module: nn.Module,
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attention_mask: Optional[torch.Tensor],
    dropout: float = 0.0,
    **kwargs,
):
    head_dim = query.shape[-1]
    scaling = 1.0 / math.sqrt(head_dim)
    attn_weights = torch.matmul(query, key.transpose(2, 3)) * scaling
    if attention_mask is not None:
        causal_mask = attention_mask[:, :, :, : key.shape[-2]]
        attn_weights = attn_weights + causal_mask

    attn_weights = nn.functional.softmax(attn_weights, dim=-1, dtype=torch.bfloat16).to(query.dtype)
    attn_weights = nn.functional.dropout(attn_weights, p=dropout, training=module.training)
    attn_output = torch.matmul(attn_weights, value)
    attn_output = attn_output.transpose(1, 2).contiguous()

    return attn_output, attn_weights


def wrapped_sageattention(
    module: nn.Module,
    query_states: torch.Tensor,
    key_states: torch.Tensor,
    value_states: torch.Tensor,
    **kwargs,
):
    assert SAGE_ATTN_AVAILABLE
    head_dim = query_states.shape[-1]
    rerange_type_head_seq = "b n s d -> b s n d"
    attn_output = sageattn(query_states, key_states, value_states)
    attn_output = rearrange(attn_output, rerange_type_head_seq, d=head_dim)
    return attn_output


def wrapped_flash_attention_3(
    module: nn.Module,
    query_states: torch.Tensor,
    key_states: torch.Tensor,
    value_states: torch.Tensor,
    last_loss=None,
    isSelfAttn=False,
    **kwargs,
):
    assert FLASH_ATTN_3_AVAILABLE
    head_dim = query_states.shape[-1]
    rerange_type_seq_head = "b n s d -> b s n d"

    q, k, v = rearrange_qkv(query_states, key_states, value_states, rerange_type_seq_head, head_dim=head_dim)

    if isSelfAttn and last_loss is not None:
        if math.isnan(last_loss):
            attn_output = flash_attn_interface.flash_attn_func(q, k, v)
        else:
            original_q = q
            original_k = k
            original_v = v
            q, qscale = symmetric_quantize(q, dtype=torch.float8_e4m3fn)
            k, kscale = symmetric_quantize(k, dtype=torch.float8_e4m3fn)
            v, vscale = symmetric_quantize(v, dtype=torch.float8_e4m3fn)
            attn_output = flash_attn_interface.flash_attn_func(
                q,
                k,
                v,
                q_descale=qscale,
                k_descale=kscale,
                v_descale=vscale,
                original_q=original_q,
                original_k=original_k,
                original_v=original_v,
            )
    else:
        attn_output = flash_attn_interface.flash_attn_func(q, k, v)

    return attn_output


def sinusoidal_embedding_1d(dim, position):
    sinusoid = torch.outer(
        position.type(torch.float64),
        torch.pow(
            10000,
            -torch.arange(dim // 2, dtype=torch.float64, device=position.device).div(dim // 2),
        ),
    )
    x = torch.cat([torch.cos(sinusoid), torch.sin(sinusoid)], dim=1)
    return x.to(position.dtype)


def precompute_freqs_cis_3d(dim: int, end: int = 1024, theta: float = 10000.0):
    # 3d rope precompute
    f_freqs_cis = precompute_freqs_cis(dim - 2 * (dim // 3), end, theta)
    h_freqs_cis = precompute_freqs_cis(dim // 3, end, theta)
    w_freqs_cis = precompute_freqs_cis(dim // 3, end, theta)
    return f_freqs_cis, h_freqs_cis, w_freqs_cis


def precompute_freqs_cis(dim: int, end: int = 1024, theta: float = 10000.0):
    # 1d rope precompute
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].double() / dim))
    freqs = torch.outer(torch.arange(end, device=freqs.device), freqs)
    freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
    return freqs_cis


def rope_apply(x, **kwargs):
    freqs = kwargs.pop("freqs")
    head_dim = kwargs.pop("head_dim")
    x = rearrange(x, "b s (n d) -> b s n d", d=head_dim)
    x_out = torch.view_as_complex(x.to(torch.float64).reshape(x.shape[0], x.shape[1], x.shape[2], -1, 2))
    x_out = torch.view_as_real(x_out * freqs).flatten(2)
    return x_out.to(x.dtype)


def pad_freqs(original_tensor, target_len):
    seq_len, s1, s2 = original_tensor.shape
    pad_size = target_len - seq_len
    padding_tensor = torch.ones(pad_size, s1, s2, dtype=original_tensor.dtype, device=original_tensor.device)
    padded_tensor = torch.cat([original_tensor, padding_tensor], dim=0)
    return padded_tensor


class RMSNorm(nn.Module):
    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(dim=-1, keepdim=True) + self.eps)

    def forward(self, x):
        dtype = x.dtype
        return self.norm(x.float()).to(dtype) * self.weight


class AttentionModule(nn.Module):
    def __init__(self, config, num_heads, head_dim):
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = head_dim
        self.attention_interface = WAN_ATTENTION_FUNCTIONS[config._attn_implementation]
        self.is_causal = False
        self.config = config

    def forward(self, query_states, key_states, value_states, **kwargs):
        query_states = rearrange(query_states, "b s (n d) -> b n s d", d=self.head_dim)
        key_states = rearrange(key_states, "b s (n d) -> b n s d", d=self.head_dim)
        value_states = rearrange(value_states, "b s (n d) -> b n s d", d=self.head_dim)

        kwargs["attention_mask"] = kwargs.get("attention_mask", None)

        deterministic = kwargs.get("deterministic", None)
        if deterministic is None:
            deterministic = os.environ.get("FLASH_ATTENTION_DETERMINISTIC", "0") == "1"
        kwargs["deterministic"] = deterministic

        attn_output = self.attention_interface(
            self,
            query_states,
            key_states,
            value_states,
            **kwargs,
        )

        if isinstance(attn_output, tuple):
            attn_output = attn_output[0]

        attn_output = rearrange(attn_output, "b s n d -> b s (n d)")
        return attn_output


class SelfAttention(nn.Module):
    def __init__(self, config, dim: int, num_heads: int, eps: float = 1e-6):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads

        self.q = nn.Linear(dim, dim)
        self.k = nn.Linear(dim, dim)
        self.v = nn.Linear(dim, dim)
        self.o = nn.Linear(dim, dim)
        self.norm_q = RMSNorm(dim, eps=eps)
        self.norm_k = RMSNorm(dim, eps=eps)

        self.attn = AttentionModule(config, self.num_heads, self.head_dim)
        self.sp_async = False

    def forward(self, x, freqs, cos, sin, last_loss):
        if not self.sp_async:
            q = self.norm_q(self.q(x))
            k = self.norm_k(self.k(x))
            v = self.v(x)

            if get_parallel_state().ulysses_enabled:
                q, k, v = gather_seq_scatter_heads_qkv(q, k, v, seq_dim=1, head_dim=2)

        else:
            q, k, v = async_ulysses_qkv_projection(
                hidden_states=x,
                seq_dimension=1,
                head_dimension=2,
                q_weight=self.q.weight,
                k_weight=self.k.weight,
                v_weight=self.v.weight,
                norm_type="rmsnorm",
                norm_q_weight=self.norm_q.weight,
                norm_k_weight=self.norm_k.weight,
                normalized_shape=self.dim,
                eps=1e-6,
                unpadded_dim_size=x.shape[1] * get_ulysses_sequence_parallel_world_size(),
            )

        q = rope_apply(q, freqs=freqs, cos=cos, sin=sin, head_dim=self.head_dim)
        k = rope_apply(k, freqs=freqs, cos=cos, sin=sin, head_dim=self.head_dim)

        x = self.attn(q, k, v, last_loss=last_loss, isSelfAttn=True)

        if not self.sp_async:
            if get_parallel_state().ulysses_enabled:
                x = gather_heads_scatter_seq(x, seq_dim=1, head_dim=2)

            x = self.o(x)
        else:
            x = async_ulysses_output_projection(
                hidden_states=x,
                seq_dimension=1,
                head_dimension=2,
                proj_weight=self.o.weight,
                proj_bias=self.o.bias,
                unpadded_dim_size=x.shape[1],
            )
        return x


class CrossAttention(nn.Module):
    def __init__(
        self,
        config,
        dim: int,
        num_heads: int,
        eps: float = 1e-6,
        has_image_input: bool = False,
    ):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads

        self.q = nn.Linear(dim, dim)
        self.k = nn.Linear(dim, dim)
        self.v = nn.Linear(dim, dim)
        self.o = nn.Linear(dim, dim)
        self.norm_q = RMSNorm(dim, eps=eps)
        self.norm_k = RMSNorm(dim, eps=eps)
        self.has_image_input = has_image_input
        if has_image_input:
            self.k_img = nn.Linear(dim, dim)
            self.v_img = nn.Linear(dim, dim)
            self.norm_k_img = RMSNorm(dim, eps=eps)

        self.attn = AttentionModule(config, self.num_heads, self.head_dim)

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        if self.has_image_input:
            img = y[:, :257]
            ctx = y[:, 257:]
        else:
            ctx = y
        q = self.norm_q(self.q(x))
        k = self.norm_k(self.k(ctx))
        v = self.v(ctx)
        x = self.attn(q, k, v)
        if self.has_image_input:
            k_img = self.norm_k_img(self.k_img(img))
            v_img = self.v_img(img)
            y = self.attn(q, k_img, v_img, head_dim=self.head_dim)
            x = x + y
        return self.o(x)


class GateModule(nn.Module):
    def __init__(
        self,
    ):
        super().__init__()

    def forward(self, x, gate, residual):
        return x + gate * residual


class DiTBlock(nn.Module):
    def __init__(
        self,
        config,
        has_image_input: bool,
        dim: int,
        num_heads: int,
        ffn_dim: int,
        eps: float = 1e-6,
    ):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.ffn_dim = ffn_dim

        self.self_attn = SelfAttention(config, dim, num_heads, eps)
        self.cross_attn = CrossAttention(config, dim, num_heads, eps, has_image_input=has_image_input)
        self.norm1 = nn.LayerNorm(dim, eps=eps, elementwise_affine=False)
        self.norm2 = nn.LayerNorm(dim, eps=eps, elementwise_affine=False)
        self.norm3 = nn.LayerNorm(dim, eps=eps)
        self.ffn = nn.Sequential(
            nn.Linear(dim, ffn_dim),
            nn.GELU(approximate="tanh"),
            nn.Linear(ffn_dim, dim),
        )
        self.modulation = nn.Parameter(torch.randn(1, 6, dim) / dim**0.5)
        self.gate = GateModule()

    def forward(self, x, context, t_mod, freqs, cos, sin, last_loss):
        shift_msa, scale_msa, gate_msa, shift_mlp, scale_mlp, gate_mlp = (
            self.modulation.to(dtype=t_mod.dtype, device=t_mod.device) + t_mod
        ).chunk(6, dim=1)
        input_x = modulate(self.norm1(x), shift_msa, scale_msa)
        x = self.gate(x, gate_msa, self.self_attn(input_x, freqs, cos, sin, last_loss))
        x = x + self.cross_attn(self.norm3(x), context)
        input_x = modulate(self.norm2(x), shift_mlp, scale_mlp)
        x = self.gate(x, gate_mlp, self.ffn(input_x))
        return x


class MLP(torch.nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.proj = torch.nn.Sequential(
            nn.LayerNorm(in_dim),
            nn.Linear(in_dim, in_dim),
            nn.GELU(),
            nn.Linear(in_dim, out_dim),
            nn.LayerNorm(out_dim),
        )

    def forward(self, x):
        return self.proj(x)


class Head(nn.Module):
    def __init__(self, dim: int, out_dim: int, patch_size: Tuple[int, int, int], eps: float):
        super().__init__()
        self.dim = dim
        self.patch_size = patch_size
        self.norm = nn.LayerNorm(dim, eps=eps, elementwise_affine=False)
        self.head = nn.Linear(dim, out_dim * math.prod(patch_size))
        self.modulation = nn.Parameter(torch.randn(1, 2, dim) / dim**0.5)

    def forward(self, x, t_mod):
        B, D = t_mod.shape
        t_mod = t_mod.view(B, 1, D)  # [B, 1, D]
        modulation = self.modulation.expand(B, -1, -1).to(t_mod)
        combined = modulation + t_mod
        shift, scale = (c.squeeze(1) for c in combined.chunk(2, dim=1))
        normalized = self.norm(x)
        modulated = normalized * (1 + scale[:, None, :]) + shift[:, None, :]
        return self.head(modulated)


class WanModel(PreTrainedModel):
    config_class = WanConfig
    _supports_flash_attn_2 = True
    supports_gradient_checkpointing = True
    _no_split_modules = ["DiTBlock"]

    def __init__(self, config: WanConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.dim = config.dim
        self.freq_dim = config.freq_dim

        # TODO: add image input
        if config.has_image_input == "true":
            self.has_image_input = True
        else:
            self.has_image_input = False
        self.patch_size = config.patch_size
        self.micro_batch_size = 1
        self.patch_embedding = nn.Conv3d(
            config.in_dim,
            config.dim,
            kernel_size=config.patch_size,
            stride=config.patch_size,
        )
        self.text_embedding = nn.Sequential(
            nn.Linear(config.text_dim, config.dim),
            nn.GELU(approximate="tanh"),
            nn.Linear(config.dim, config.dim),
        )
        self.time_embedding = nn.Sequential(
            nn.Linear(config.freq_dim, config.dim),
            nn.SiLU(),
            nn.Linear(config.dim, config.dim),
        )
        self.time_projection = nn.Sequential(
            nn.SiLU(),
            nn.Linear(config.dim, config.dim * 6),
        )
        self.blocks = nn.ModuleList(
            [
                DiTBlock(
                    config,
                    self.has_image_input,
                    config.dim,
                    config.num_heads,
                    config.ffn_dim,
                    config.eps,
                )
                for _ in range(config.num_layers)
            ]
        )
        self.head = Head(config.dim, config.out_dim, config.patch_size, config.eps)
        head_dim = config.dim // config.num_heads
        self.freqs = precompute_freqs_cis_3d(head_dim)

        if self.has_image_input:
            self.img_emb = MLP(1280, config.dim)  # clip_feature_dim = 1280

        self.gradient_checkpointing = False

    def patchify(self, x: torch.Tensor):
        x = self.patch_embedding(x)
        grid_size = x.shape[2:]
        x = rearrange(x, "b c f h w -> b (f h w) c").contiguous()
        return x, grid_size  # x, grid_size: (f, h, w)

    def unpatchify(self, x: torch.Tensor, grid_size: torch.Tensor):
        return rearrange(
            x,
            "b (f h w) (x y z c) -> b c (f x) (h y) (w z)",
            f=grid_size[0],
            h=grid_size[1],
            w=grid_size[2],
            x=self.patch_size[0],
            y=self.patch_size[1],
            z=self.patch_size[2],
        )

    def forward(
        self,
        x: torch.Tensor,
        timestep: torch.Tensor,
        context: torch.Tensor,
        clip_feature: Optional[torch.Tensor] = None,
        y: Optional[torch.Tensor] = None,
        last_loss=None,
        **kwargs,
    ):
        t = self.time_embedding(sinusoidal_embedding_1d(self.freq_dim, timestep))

        t_mod = self.time_projection(t).unflatten(1, (6, self.dim))
        context = self.text_embedding(context)

        if self.has_image_input:
            if self.micro_batch_size > 1:
                x = torch.cat([x, y], dim=1)  # (b, c_x + c_y, f, h, w)
                clip_embdding = self.img_emb(clip_feature)
                context = context.squeeze(1)
                clip_embdding = clip_embdding.squeeze(1)
                context = torch.cat([clip_embdding, context], dim=1)
            else:
                x = torch.cat([x, y], dim=1)  # (b, c_x + c_y, f, h, w)
                clip_embdding = self.img_emb(clip_feature)
                context = torch.cat([clip_embdding, context], dim=1)

        x, (f, h, w) = self.patchify(x)

        freqs = (
            torch.cat(
                [
                    self.freqs[0][:f].view(f, 1, 1, -1).expand(f, h, w, -1),
                    self.freqs[1][:h].view(1, h, 1, -1).expand(f, h, w, -1),
                    self.freqs[2][:w].view(1, 1, w, -1).expand(f, h, w, -1),
                ],
                dim=-1,
            )
            .reshape(f * h * w, 1, -1)
            .to(x.device)
        )
        cos = freqs.real.squeeze().contiguous()
        sin = freqs.imag.squeeze().contiguous()

        if get_parallel_state().ulysses_enabled:
            x = slice_input_tensor_scale_grad(x, dim=1)

        for block in self.blocks:
            if self.training and self.gradient_checkpointing:
                x = self._gradient_checkpointing_func(
                    block.__call__,
                    x,
                    context,
                    t_mod,
                    freqs,
                    cos,
                    sin,
                    last_loss=last_loss,
                )
            else:
                x = block(x, context, t_mod, freqs, cos, sin, last_loss=last_loss)

        x = self.head(x, t)

        if get_parallel_state().ulysses_enabled:
            x = gather_outputs(x, gather_dim=1)
        x = self.unpatchify(x, (f, h, w))
        return x


if is_liger_kernel_available():
    RMSNorm = LigerRMSNorm
    logger.info_rank0("Apply liger kernel to Wan.")

try:
    from veomni.ops.dit.rope_wan.rotary import apply_rotary_emb

    rope_apply = apply_rotary_emb
    logger.info_rank0("Apply fused interleaved rope to Wan.")
except ImportError:
    pass

WAN_ATTENTION_FUNCTIONS: Dict[str, Dict[str, Callable]] = {}
WAN_ATTENTION_FUNCTIONS.update(ALL_ATTENTION_FUNCTIONS)
WAN_ATTENTION_FUNCTIONS.update(
    {
        "eager": eager_attention_forward,
        "flash_attention_3": wrapped_flash_attention_3,
        "sageattention": wrapped_sageattention,
    }
)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
import re
import time
from collections import OrderedDict
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Literal, Optional, Sequence, Tuple, Union

import torch


try:
    from hdfs_io import copy  # for internal use only
except ImportError:
    from ..utils.hdfs_io import copy
from diffusers.utils import SAFE_WEIGHTS_INDEX_NAME as DIFFUSERS_SAFE_WEIGHTS_INDEX_NAME
from diffusers.utils import SAFETENSORS_WEIGHTS_NAME as DIFFUSERS_SAFETENSORS_WEIGHTS_NAME
from torch import distributed as dist
from torch import nn
from tqdm import tqdm
from transformers.utils import SAFE_WEIGHTS_INDEX_NAME, SAFE_WEIGHTS_NAME, WEIGHTS_INDEX_NAME, WEIGHTS_NAME
from transformers.utils.hub import cached_file, get_checkpoint_shard_files
from transformers.utils.import_utils import is_safetensors_available

from ..distributed.parallel_state import get_parallel_state
from ..utils import logging
from ..utils.device import synchronize
from ..utils.helper import empty_cache, get_cache_dir, get_dtype_size


if is_safetensors_available():
    from safetensors import safe_open
    from safetensors.torch import save_file


if TYPE_CHECKING:
    from transformers import GenerationConfig, PretrainedConfig, PreTrainedModel, PreTrainedTokenizer, ProcessorMixin

    from ..distributed.parallel_plan import ParallelPlan

    ModelAssets = Union[GenerationConfig, PretrainedConfig, PreTrainedTokenizer, ProcessorMixin]


logger = logging.get_logger(__name__)


@contextmanager
def init_empty_weights():
    """
    A context manager under which models are initialized with all parameters on the meta device.

    Borrowed from: https://github.com/huggingface/accelerate/blob/v1.0.0rc1/src/accelerate/big_modeling.py#L57
    """
    old_register_parameter = nn.Module.register_parameter

    def register_empty_parameter(module: "nn.Module", name: str, param: "nn.Parameter"):
        old_register_parameter(module, name, param)
        if param is not None:
            param_cls = type(module._parameters[name])
            kwargs = module._parameters[name].__dict__
            kwargs["requires_grad"] = param.requires_grad
            # When we have a case of tensor2 = tensor1, it would call the set_attr
            # of param, which in turn would call the register_parameter API.
            # In this case, the new param is already on meta-device, since it was moved
            # previously when it was initialized. Hence, when resetting, you can
            # directly assign that tensor instead of re-init. If you re-init you would
            # lose the relationship.
            module._parameters[name] = (
                param
                if param.device == torch.device("meta")
                else param_cls(module._parameters[name].to("meta"), **kwargs)
            )

    try:
        nn.Module.register_parameter = register_empty_parameter
        yield
    finally:
        nn.Module.register_parameter = old_register_parameter


@dataclass
class StateDictIterator:
    filepath: str

    def __iter__(self) -> Generator[Tuple[str, "torch.Tensor"], None, None]:
        if self.filepath.endswith(".safetensors"):
            with safe_open(self.filepath, framework="pt", device="cpu") as f:
                for key in f.keys():
                    yield key, f.get_tensor(key)

        else:
            state_dict = torch.load(self.filepath, map_location="cpu", weights_only=True, mmap=True)
            for key in state_dict.keys():
                yield key, state_dict[key]


@dataclass
class BroadcastMetadata:
    done: bool
    name: Optional[str]
    shape: Optional["torch.Size"]
    dtype: Optional["torch.dtype"]


def _load_state_dict(weights_path: str, **kwargs) -> List["StateDictIterator"]:
    """
    Loads (sharded) state dict in transformers' format.
    """
    cache_kwargs = {"_raise_exceptions_for_missing_entries": False, **kwargs}
    resolved_weight_file = cached_file(weights_path, SAFE_WEIGHTS_NAME, **cache_kwargs)
    if resolved_weight_file:
        return [StateDictIterator(resolved_weight_file)]

    resolved_weight_file = cached_file(weights_path, SAFE_WEIGHTS_INDEX_NAME, **cache_kwargs)
    if resolved_weight_file:
        shard_files, _ = get_checkpoint_shard_files(weights_path, resolved_weight_file, **kwargs)
        return [StateDictIterator(shard_file) for shard_file in shard_files]

    resolved_weight_file = cached_file(weights_path, DIFFUSERS_SAFETENSORS_WEIGHTS_NAME, **cache_kwargs)
    if resolved_weight_file:
        return [StateDictIterator(resolved_weight_file)]

    resolved_weight_file = cached_file(weights_path, DIFFUSERS_SAFE_WEIGHTS_INDEX_NAME, **cache_kwargs)
    if resolved_weight_file:
        shard_files, _ = get_checkpoint_shard_files(weights_path, resolved_weight_file, **kwargs)
        return [StateDictIterator(shard_file) for shard_file in shard_files]

    resolved_weight_file = cached_file(weights_path, WEIGHTS_NAME, **cache_kwargs)
    if resolved_weight_file:
        return [StateDictIterator(resolved_weight_file)]

    resolved_weight_file = cached_file(weights_path, WEIGHTS_INDEX_NAME, **cache_kwargs)
    if resolved_weight_file:
        shard_files, _ = get_checkpoint_shard_files(weights_path, resolved_weight_file, **kwargs)
        return [StateDictIterator(shard_file) for shard_file in shard_files]

    raise ValueError(f"Cannot find checkpoint files in {weights_path}.")


def _find_submodule(module: "nn.Module", name: str) -> Tuple["nn.Module", str]:
    """
    Finds the leaf module according to the name.
    """
    pieces = name.split(".")
    for piece in pieces[:-1]:
        if not hasattr(module, piece):
            raise ValueError(f"Cannot find {piece} in {module}.")

        module = getattr(module, piece)

    return module, pieces[-1]


def _dispatch_parameter(
    module: "nn.Module",
    name: str,
    tensor: "torch.Tensor",
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
    parallel_plan: Optional["ParallelPlan"] = None,
) -> None:
    """
    Assigns parameter to an empty model.

    NOTE: FSDP module must use in-place operators.
    """
    full_param_name = name
    module, local_name = _find_submodule(module, name)
    orig_tensor = module._parameters[local_name].data

    # Handle parameter slicing according to parallel_plan, now only EP-aware
    if parallel_plan is not None:
        tensor = parallel_plan.shard_tensor(tensor, full_param_name, orig_tensor.shape)

    tensor = tensor.to(orig_tensor)
    if hasattr(orig_tensor, "device_mesh"):  # dtensor
        if orig_tensor.device.type == "cpu":
            raise ValueError("Cannot load dtensor on CPU.")

        device_mesh = getattr(orig_tensor, "device_mesh")
        placements = getattr(orig_tensor, "placements")
        module._parameters[local_name].data.copy_(dtensor_factory(tensor, device_mesh, placements))
    else:  # not dtensor
        module._parameters[local_name].data.copy_(tensor)


def _dispatch_buffer(
    module: "nn.Module",
    name: str,
    buffer: "torch.Tensor",
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
) -> None:
    """
    Assigns buffer to an empty model.
    """
    module, name = _find_submodule(module, name)
    orig_tensor = module._buffers[name]

    if hasattr(orig_tensor, "device_mesh"):  # dtensor buffer
        if dtensor_factory is None:
            raise ValueError("dtensor buffer requires a dtensor_factory.")

        device_mesh = getattr(orig_tensor, "device_mesh")
        placements = getattr(orig_tensor, "placements")
        module._buffers[name] = dtensor_factory(buffer.to(dtype=orig_tensor.dtype), device_mesh, placements)
    else:
        module._buffers[name].copy_(buffer.to(device=orig_tensor.device, dtype=orig_tensor.dtype))


def _init_parameter(
    module: "nn.Module",
    name: str,
) -> None:
    """
    Initializes parameter in model.
    """
    pieces = name.split(".")
    init_func = None
    for piece in pieces[:-1]:
        if not hasattr(module, piece):
            raise ValueError(f"Cannot find {piece} in {module}.")

        if hasattr(module, "_init_weights"):
            init_func = getattr(module, "_init_weights")

        module = getattr(module, piece)

    if init_func is None:
        raise ValueError(f"Cannot retrieve `_init_weights` function in the parents of {module}.")

    module.apply(init_func)


def _convert_weight_key(key: str, model: "PreTrainedModel") -> str:
    """
    Convert a single state dict key using the model's checkpoint conversion mapping.

    For example, in the InternVL, we have _checkpoint_conversion_mapping = {"^model": "language_model"}

    This is to adapt to the big breaking change introduced in HF transformers 4.52:
    https://github.com/huggingface/transformers/pull/38385
    """
    if not hasattr(model, "_checkpoint_conversion_mapping"):
        return key

    for pattern, replacement in model._checkpoint_conversion_mapping.items():
        replacement = replacement.lstrip("^")  # strip off un-needed chars and patterns
        replacement = re.sub(r"\(.*\)", "", replacement)
        converted_key, n_replace = re.subn(pattern, replacement, key)
        # Early exit of the loop
        if n_replace > 0:
            return converted_key

    return key


@torch.no_grad()
def load_model_weights(
    model: Union["nn.Module", "PreTrainedModel"],
    weights_path: str,
    init_device: Literal["cpu", "cuda", "npu"] = "cuda",
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
) -> None:
    """
    Loads pre-trained model states in transformers' format.
    """
    buffer_dict = {name: buffer.clone() for name, buffer in model.named_buffers()}
    parameter_names_to_load = {name for name, _ in model.named_parameters()}
    model.to_empty(device=init_device)

    # Get parallel plan if available
    parallel_plan = None
    if hasattr(model, "get_parallel_plan"):
        parallel_plan = model.get_parallel_plan()

    state_dict_iterators = _load_state_dict(weights_path)
    for state_dict_iterator in tqdm(
        state_dict_iterators, desc="Loading checkpoint shards", disable=int(os.getenv("LOCAL_RANK", "-1")) > 0
    ):
        for name, tensor in state_dict_iterator:
            # IMPORTANT: Call this function to adapt to transformers 4.52 breaking change
            # on model structure. See the comment for details.
            name = _convert_weight_key(name, model)

            if name in buffer_dict.keys():  # persistent buffers
                buffer_dict[name] = tensor.clone()
            elif name in parameter_names_to_load:
                parameter_names_to_load.remove(name)
                _dispatch_parameter(model, name, tensor, dtensor_factory, parallel_plan)
            else:
                logger.info_rank0(f"Unexpected key in state dict: {name}.")

        del state_dict_iterator
        empty_cache()

    post_process_after_weight_loading(model, buffer_dict, parameter_names_to_load, dtensor_factory)


@torch.no_grad()
def rank0_load_and_broadcast_weights(
    model: Union["nn.Module", "PreTrainedModel"],
    weights_path: str,
    init_device: Literal["cpu", "cuda", "npu"] = "cuda",
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
):
    """
    This functions serves as the same purpose as `load_model_weights`
    but reduces disk I/O by broadcasting weights from rank0.
    In comparison, `load_model_weights` would require every GPU to go through the entire model weights on disk.
    """
    if not dist.is_available() or not dist.is_initialized():
        logger.warning_once("Distributed environment not initialized, falling back to load_model_weights.")
        return load_model_weights(model, weights_path, init_device, dtensor_factory)

    buffer_dict = {name: buffer.clone() for name, buffer in model.named_buffers()}
    parameter_names_to_load = {name for name, _ in model.named_parameters()}
    model.to_empty(device=init_device)

    # Get parallel plan if available
    parallel_plan = None
    if hasattr(model, "get_parallel_plan"):
        parallel_plan = model.get_parallel_plan()

    global_rank = get_parallel_state().global_rank
    torch_device = torch.device(init_device)

    # get the safetensor file iterator
    state_dict_iterators = _load_state_dict(weights_path) if global_rank == 0 else None
    shard_count = len(state_dict_iterators) if global_rank == 0 else 0
    logger.info_rank0(f"rank0_load_and_broadcast_weights: {shard_count=} ")
    shard_count_tensor = torch.tensor(
        [shard_count],
        dtype=torch.int64,
        device=torch_device if torch_device.type != "cpu" else torch.device("cpu"),
    )
    dist.broadcast(shard_count_tensor, src=0)
    shard_count = int(shard_count_tensor.item())

    if global_rank == 0:
        # only rank0 would actual read weights from safetensor state_dict iterators
        shard_iterable = enumerate(
            tqdm(
                state_dict_iterators,
                desc="Loading checkpoint shards",
                # only rank0 displays tqdm pbar
                disable=int(os.getenv("LOCAL_RANK", "-1")) > 0,
            )
        )
    else:
        shard_iterable = enumerate(range(shard_count))

    # iterate safetensor files; each file would have a iterator to read weight keys and tensors
    for shard_idx, shard_payload in shard_iterable:
        state_dict_iterator = shard_payload if global_rank == 0 else None
        iterator = iter(state_dict_iterator) if global_rank == 0 else None

        while True:
            # read tensors from safetensor
            tensor: Optional["torch.Tensor"] = None

            if global_rank == 0:
                try:
                    key, tensor = next(iterator)  # type: ignore[arg-type]
                    key = _convert_weight_key(key, model)
                    logger.info_rank0(f"loading {key=}")
                    if torch.count_nonzero(tensor) == 0:
                        logger.warning_rank0(f"Detected tensor with all-zero values when reading safetensor: {key=}")
                    metadata = BroadcastMetadata(False, key, tensor.shape, tensor.dtype)
                except StopIteration:
                    metadata = BroadcastMetadata(True, None, None, None)
            else:
                metadata = BroadcastMetadata(False, None, None, None)

            metadata_list = [metadata]
            dist.broadcast_object_list(metadata_list, src=0)
            metadata = metadata_list[0]

            if metadata.done:
                break

            name = metadata.name
            shape = metadata.shape
            dtype = metadata.dtype
            if name is None or shape is None or dtype is None:
                raise RuntimeError("Received incomplete broadcast metadata.")
            logger.info_rank0(f"rank0_load_and_broadcast_weights: broadcasting {name=}")
            if global_rank != 0:
                tensor = torch.empty(shape, dtype=dtype, device=torch_device)
            else:
                tensor = tensor.to(torch_device, non_blocking=True)  # type: ignore[assignment]

            start_time = time.perf_counter()
            dist.broadcast(tensor, src=0)
            logger.info_rank0(
                f"{name=}, {shape=}, {dtype=}, broadcast time (ms) spent: {1000 * (time.perf_counter() - start_time)}"
            )

            if name in buffer_dict:
                buffer_dict[name] = tensor.detach().clone()
            elif name in parameter_names_to_load:
                parameter_names_to_load.discard(name)
                _dispatch_parameter(model, name, tensor, dtensor_factory, parallel_plan)
            else:
                if global_rank == 0:
                    logger.info_rank0(f"Unexpected key in state dict: {name}.")

            del tensor

        if global_rank == 0:
            del state_dict_iterator

        empty_cache()

    post_process_after_weight_loading(model, buffer_dict, parameter_names_to_load, dtensor_factory)


def post_process_after_weight_loading(
    model: Union["nn.Module", "PreTrainedModel"],
    buffer_dict,
    parameter_names_left: Optional[set[str]] = None,
    dtensor_factory: Optional[Callable[["torch.Tensor", Any, Any], "torch.Tensor"]] = None,
):
    """
    shared logic after weight loading that handles buffer, missing weight keys and tied embedding weights.
    """
    parameter_names_left = parameter_names_left or set()

    for name, buffer in buffer_dict.items():
        _dispatch_buffer(model, name, buffer, dtensor_factory)

    if parameter_names_left:
        logger.info_rank0(f"Find missing key(s) in state dict: {parameter_names_left}, initialize them.")
        for name in parameter_names_left:
            _init_parameter(model, name)

    # we should tie embeddings after loading weights because to_empty() leads to untied weights,
    # except for fsdp1 (custom init) and fsdp2 (swap tensor) contexts.
    if getattr(model.config, "tie_word_embeddings", True):
        try:
            input_embeddings = model.get_input_embeddings()
            output_embeddings = model.get_output_embeddings()
            output_embeddings._parameters["weight"] = input_embeddings._parameters["weight"]
        except Exception as e:
            logger.info_rank0(f"Failed to tie embeddings: {e}")
            raise RuntimeError("Failed to tie input/output embeddings") from e


def _get_shard_info(
    state_dict: Dict[str, "torch.Tensor"],
    save_dtype: Optional[Union[str, "torch.dtype"]],
    shard_size: int,
    safe_serialization: bool,
) -> Tuple[bool, int, Dict[str, str]]:
    """
    Gets the shard information, should be executed at rank 0.
    """
    current_size, total_size = 0, 0
    current_shard, shard_list = [], []
    for name, tensor in state_dict.items():
        if isinstance(save_dtype, str):
            dtype = getattr(torch, save_dtype)
        elif isinstance(save_dtype, torch.dtype):
            dtype = save_dtype
        else:
            dtype = tensor.dtype
        tensor_size = tensor.numel() * get_dtype_size(dtype)  # dtensor's numel == tensor's numel
        if current_size != 0 and current_size + tensor_size > shard_size:
            total_size += current_size
            shard_list.append(current_shard)
            current_size = 0
            current_shard = []

        current_size += tensor_size
        current_shard.append(name)

    if current_size != 0:
        total_size += current_size
        shard_list.append(current_shard)

    weights_name = SAFE_WEIGHTS_NAME if safe_serialization else WEIGHTS_NAME
    num_shards = len(shard_list)
    weight_map = OrderedDict()
    is_sharded = None
    if num_shards == 1:
        is_sharded = False
        for name in shard_list[0]:
            weight_map[name] = weights_name
    else:
        is_sharded = True
        for shard_idx, shard in enumerate(shard_list):
            prefix, extension = weights_name.rsplit(".", maxsplit=1)
            file_name = f"{prefix}-{shard_idx + 1:05d}-of-{num_shards:05d}.{extension}"
            for name in shard:
                weight_map[name] = file_name

    return is_sharded, total_size, weight_map


def _save_state_dict(
    state_dict: Dict[str, "torch.Tensor"],
    path_to_save: "os.PathLike",
    safe_serialization: bool,
) -> None:
    """
    Save function.
    """
    if safe_serialization:
        save_file(state_dict, path_to_save, metadata={"format": "pt"})
    else:
        torch.save(state_dict, path_to_save)


@torch.no_grad()
def save_model_weights(
    output_dir: Union[str, "os.PathLike"],
    state_dict: Dict[str, "torch.Tensor"],
    global_rank: Optional[int] = None,
    save_dtype: Optional[Union[str, "torch.dtype"]] = "bfloat16",
    shard_size: int = 5_000_000_000,
    safe_serialization: bool = True,
    model_assets: Optional[Sequence["ModelAssets"]] = None,
) -> None:
    """
    Saves full model weights. The model parameters should be either tensor or dtensor.

    If global_rank is given, it will assume it is executed on all ranks.
    """
    if output_dir.startswith("hdfs://"):
        hdfs_dir = output_dir
        hdfs_upper_dir = output_dir.rstrip("/")
        hdfs_upper_dir = hdfs_upper_dir[: hdfs_upper_dir.rfind("/")]
        output_dir = get_cache_dir(output_dir)
    else:
        hdfs_dir = None

    os.makedirs(output_dir, exist_ok=True)
    is_sharded, total_size, weight_map = _get_shard_info(state_dict, save_dtype, shard_size, safe_serialization)
    full_state_dict = OrderedDict()
    prev_file_name = None
    for name, tensor in state_dict.items():
        if hasattr(tensor.data, "full_tensor"):  # dtensor
            tensor = tensor.data.full_tensor()
        else:
            tensor = tensor.data

        if save_dtype:
            tensor = tensor.to(dtype=getattr(torch, save_dtype) if isinstance(save_dtype, str) else save_dtype)

        if prev_file_name is not None and weight_map[name] != prev_file_name:
            if global_rank is None or global_rank == 0:
                _save_state_dict(full_state_dict, os.path.join(output_dir, prev_file_name), safe_serialization)
                full_state_dict = OrderedDict()

            empty_cache()
            if global_rank is not None and dist.is_initialized():  # avoid process hanging
                synchronize()
                dist.barrier()

        if global_rank is None or global_rank == 0:
            full_state_dict[name] = tensor.detach().cpu()

        prev_file_name = weight_map[name]
        del tensor

    if global_rank is None or global_rank == 0:
        if len(full_state_dict):
            _save_state_dict(full_state_dict, os.path.join(output_dir, prev_file_name), safe_serialization)

        if is_sharded:
            index = {
                "metadata": {"total_size": total_size},
                "weight_map": weight_map,
            }
            index_file = SAFE_WEIGHTS_INDEX_NAME if safe_serialization else WEIGHTS_INDEX_NAME
            with open(os.path.join(output_dir, index_file), "w", encoding="utf-8") as f:
                content = json.dumps(index, indent=2, sort_keys=True) + "\n"
                f.write(content)

            logger.info(f"Model weight splits saved in {output_dir}.")
        else:
            logger.info(f"Model weights saved at {os.path.join(output_dir, prev_file_name)}.")

        if model_assets is not None:
            for model_asset in model_assets:
                if hasattr(model_asset, "save_pretrained"):
                    model_asset.save_pretrained(output_dir)
                else:
                    logger.warning(f"Model asset {model_asset} should implement `save_pretrained`.")

        if hdfs_dir is not None:
            copy(output_dir, hdfs_upper_dir)
            logger.info(f"Model weights uploaded to {hdfs_dir}.")


def save_model_assets(output_dir: Union[str, "os.PathLike"], model_assets: Sequence["ModelAssets"]):
    if output_dir.startswith("hdfs://"):
        hdfs_dir = output_dir
        hdfs_upper_dir = output_dir.rstrip("/")
        hdfs_upper_dir = hdfs_upper_dir[: hdfs_upper_dir.rfind("/")]
        output_dir = get_cache_dir(output_dir)
    else:
        hdfs_dir = None

    for model_asset in model_assets:
        if hasattr(model_asset, "save_pretrained"):
            model_asset.save_pretrained(output_dir)
        else:
            logger.warning(f"Model asset {model_asset} should implement `save_pretrained`.")

    if hdfs_dir is not None:
        copy(output_dir, hdfs_upper_dir)
        logger.info(f"Model config and tokenizer uploaded to {hdfs_dir}.")


class GradientCheckpointingLayer(nn.Module):
    """Base class for layers with gradient checkpointing.

    This class enables gradient checkpointing functionality for a layer. By default, gradient checkpointing is disabled
    (`gradient_checkpointing = False`). When `model.set_gradient_checkpointing()` is called, gradient checkpointing is
    enabled by setting `gradient_checkpointing = True` and assigning a checkpointing function to `_gradient_checkpointing_func`.

    Important:

        When using gradient checkpointing with `use_reentrant=True`, inputs that require gradients (e.g. hidden states)
        must be passed as positional arguments (`*args`) rather than keyword arguments to properly propagate gradients.

        Example:

            ```python
            >>> # Correct - hidden_states passed as positional arg
            >>> out = self.layer(hidden_states, attention_mask=attention_mask)

            >>> # Incorrect - hidden_states passed as keyword arg
            >>> out = self.layer(hidden_states=hidden_states, attention_mask=attention_mask)
            ```
    """

    gradient_checkpointing = False

    def __call__(self, *args, **kwargs):
        if self.gradient_checkpointing and self.training:
            return self._gradient_checkpointing_func(partial(super().__call__, **kwargs), *args)
        return super().__call__(*args, **kwargs)



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import custom, seed_omni, transformers
from .auto import build_foundation_model, build_processor, build_tokenizer
from .module_utils import (
    init_empty_weights,
    load_model_weights,
    rank0_load_and_broadcast_weights,
    save_model_assets,
    save_model_weights,
)


__all__ = [
    "build_foundation_model",
    "build_processor",
    "build_tokenizer",
    "init_empty_weights",
    "load_model_weights",
    "rank0_load_and_broadcast_weights",
    "save_model_assets",
    "save_model_weights",
    "transformers",
    "seed_omni",
    "custom",
]



from collections import defaultdict
from typing import List, Union

import torch
import torch.distributed as dist

from veomni.utils.dist_utils import all_reduce

from ...data.constants import IGNORE_INDEX
from ...utils.device import get_device_type
from .modeling_seed_omni import SeedOmniModel


def omni_token_meter(batches: Union[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]):
    """Calculate the total number of text_tokens/image_tokens in a global batch, or one micro batch."""
    if isinstance(batches, dict):
        batches = [batches]
    token_meter = defaultdict(int)
    for batch in batches:
        token_meter["text_tokens"] += torch.sum(batch["labels"] != IGNORE_INDEX)
        if "image_output_mask" in batch:
            token_meter["image_tokens"] += torch.sum(batch["image_output_mask"])
    return token_meter


def mean_global_batch_loss(
    losses: dict[str, int], micro_batch_meter: dict[str, torch.Tensor], micro_batches_meter: dict[str, int]
):
    """Calculate the mean loss in a global batch. Avg on token_num instead of on len(global_batch).
    For text_loss:
    - √ text_loss = text_loss * micro_batch_text_token_num / global_batch_text_token_num
    - x text_loss = text_loss / (len(global_batch))
    For image loss:
    - √ image_loss = image_loss * micro_batch_image_token_num / global_batch_image_token_num
    - x image_loss = image_loss / (len(global_batch))
    """
    loss = torch.tensor(0.0, device=get_device_type())
    for key in losses.keys():
        if "foundation" in key:
            losses[key] *= micro_batch_meter["text_tokens"]
            if micro_batches_meter["text_tokens"] != 0:
                losses[key] /= micro_batches_meter["text_tokens"]

        elif "image_decoder" in key:
            losses[key] *= micro_batch_meter["image_tokens"]
            if micro_batches_meter["image_tokens"] != 0:
                losses[key] /= micro_batches_meter["image_tokens"]
        else:
            raise ValueError(f"Unrecognized loss key: {key}")
        loss += losses[key]
    return loss


def mean_dp_loss(
    losses: dict[str, int], micro_batches_meter: dict[str, torch.Tensor], group: dist.ProcessGroup = None
):
    """Calcuate the mean loss in a dp group. Avg on all_reduced_token_num instead of on dp_size.
    For text_loss:
    - √ text_loss = all_reduce(text_loss * micro_batch_text_token_num, op="sum") / all_reduce(global_batch_text_token_num, op="sum")
    - x text_loss = all_reduce(text_loss, op="sum") / dp_size
    For image loss:
    - √ image_loss = all_reduce(image_loss * micro_batch_image_token_num, op="sum") / all_reduce(global_batch_image_token_num, op="sum")
    - x image_loss = all_reduce(image_loss, op="sum") / dp_size
    """
    total_loss = 0
    total_losses = {}
    for key, item in losses.items():
        if "foundation" in key:
            item *= micro_batches_meter["text_tokens"]
            all_reduced_loss = all_reduce((item.item()), op="sum", group=group)
            all_reduced_length = all_reduce((micro_batches_meter["text_tokens"].item()), op="sum", group=group)
            if all_reduced_length != 0:
                all_reduced_loss /= all_reduced_length
            total_losses[key] = all_reduced_loss
            total_loss += all_reduced_loss
        elif "image_decoder" in key:
            item *= micro_batches_meter["image_tokens"]
            all_reduced_loss = all_reduce((item.item()), op="sum", group=group)
            all_reduced_length = all_reduced_length = all_reduce(
                (micro_batches_meter["image_tokens"].item()), op="sum", group=group
            )
            if all_reduced_length != 0:
                all_reduced_loss /= all_reduced_length
            total_losses[key] = all_reduced_loss
            total_loss += all_reduced_loss
        else:
            raise ValueError(f"Unrecognized loss key: {key}")
    return total_loss, total_losses


def get_embed_token_grad_mask_func(
    model: SeedOmniModel, trained_embedding_indices: List[int], global_rank: int, world_size: int
):
    embed_dim = model.encoder.text_encoder.weight.shape[-1]
    train_indices = []
    for indice in trained_embedding_indices:
        train_indices.append((indice * embed_dim, indice * embed_dim + embed_dim))

    def mask_embedding_grad(embedding: torch.nn.Embedding):
        grad = embedding.weight.grad
        device = embedding.weight.device
        if grad is None:
            local_numel = torch.tensor([0], device=device)
        else:
            local_numel = torch.tensor([grad.numel()], device=device)

        all_numel = [torch.zeros_like(local_numel) for _ in range(world_size)]
        dist.all_gather(all_numel, local_numel)

        if grad is None:
            return
        shard_sizes = [x.item() for x in all_numel]
        offsets = [0]
        for sz in shard_sizes:
            offsets.append(offsets[-1] + sz)

        start_offset = offsets[global_rank]
        end_offset = offsets[global_rank + 1]

        local_mask = torch.zeros_like(grad, dtype=torch.bool, device=device)
        for start, end in train_indices:
            if end >= start_offset and start < end_offset:
                local_mask[max(start, start_offset) - start_offset : min(end, end_offset) - start_offset] = True
        grad.mul_(local_mask)

    return mask_embedding_grad



from typing import Dict

import torch
import torch.nn as nn

from ....transformers.qwen2_5vl.modeling_qwen2_5_vl import Qwen2_5_VisionTransformerPretrainedModel
from ...projector import build_feature_projector
from ..base import BaseEncoderModelMixin
from .configuration_qwen2_5_vl_vision_model import Qwen25VLVisionModelConfig


class Qwen25VLVisionModel(BaseEncoderModelMixin, Qwen2_5_VisionTransformerPretrainedModel):
    config_class = Qwen25VLVisionModelConfig
    _no_split_modules = ["Qwen2_5_VLVisionBlock"]

    def __init__(self, config: Qwen25VLVisionModelConfig):
        super().__init__(config)
        self.config = config
        if config.add_projector and config.output_size is not None:
            self.projector = build_feature_projector(config.out_hidden_size, config.output_size)
        else:
            if config.output_size and config.output_size != config.out_hidden_size:
                raise ValueError("`output_size` should be same as `out_hidden_size`.")

            self.projector = nn.Identity()

    def set_projector_trainable_only(self):
        self.requires_grad_(False)
        if self.config.add_projector and self.config.output_size is not None:
            self.projector.requires_grad_(True)
            if self.config.train_origin_projector:
                self.merger.requires_grad_(True)
        else:
            self.merger.requires_grad_(True)

    def lm_encode(self, features: torch.Tensor, grid_thw: torch.Tensor, **kwargs) -> torch.Tensor:
        return super().forward(features, grid_thw)

    def _get_lm_dummy_data(self) -> Dict[str, torch.Tensor]:
        pixel_values = torch.randn((4, 3 * 2 * 14 * 14), dtype=self.dtype, device=self.device)
        grid_thw = torch.tensor([[1, 2, 2]], dtype=torch.int32, device=self.device)
        return {"features": pixel_values, "grid_thw": grid_thw}



from typing import Literal, Optional, Tuple

import torch
from transformers.modeling_flash_attention_utils import _flash_attention_forward

from ..distributed.parallel_state import get_parallel_state
from ..distributed.sequence_parallel import (
    gather_heads_scatter_seq,
    gather_seq_scatter_heads,
)
from ..utils import logging
from ..utils.import_utils import is_seed_kernels_available


if is_seed_kernels_available():
    from seed_kernels.transformers.functional import seed_flash_attention_forward

logger = logging.get_logger(__name__)


def flash_attention_forward(
    module: torch.nn.Module,
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attention_mask: Optional[torch.Tensor],
    dropout: float = 0.0,
    scaling: Optional[float] = None,
    sliding_window: Optional[int] = None,
    softcap: Optional[float] = None,
    implementation: Optional[Literal["fa2", "lego", "fa3"]] = None,
    skip_ulysses: bool = False,  # Skip ulysses for some ViT cases like internvl3.5
    **kwargs,
) -> Tuple[torch.Tensor, None]:
    if kwargs.get("output_attentions", False) or kwargs.get("head_mask", None) is not None:
        logger.warning_once(
            "`flash_attention_2` does not support `output_attentions=True` or `head_mask`."
            " Please set your attention to `eager` if you want any of these features."
        )

    # FA2 uses non-transposed inputs
    query = query.transpose(1, 2)
    key = key.transpose(1, 2)
    value = value.transpose(1, 2)

    # FA2 always relies on the value set in the module, so remove it if present in kwargs to avoid passing it twice
    kwargs.pop("is_causal", None)

    # This is for Qwen2VL's mrope
    position_ids = kwargs.pop("position_ids", None)
    if position_ids is not None and position_ids.dim() == 3:
        position_ids = position_ids[0]

    # Ulysses patch
    ulysses_enabled = get_parallel_state().ulysses_enabled
    if ulysses_enabled and not skip_ulysses:
        ulysses_group = get_parallel_state().ulysses_group
        # Sanity Check & Repeat Key & Value
        ulysses_size = get_parallel_state().ulysses_size
        q_head_num = query.shape[2]
        kv_head_num = key.shape[2]
        unpadded_seq_len = None

        assert q_head_num % ulysses_size == 0, (
            f"num_query_heads ({q_head_num}) must be divisible by ulysses_size ({ulysses_size})"
        )
        if ulysses_size > kv_head_num:
            assert ulysses_size % kv_head_num == 0, (
                f"ulysses_size ({ulysses_size}) must be divisible by num_key_value_heads ({kv_head_num})"
            )
            n_repeat = ulysses_size // kv_head_num
            # Shape before: (batch_size, seq_len, kv_head_num, head_dim)
            # This repeats the K/V heads (dim 2) to match the ulysses_size (SP world size)
            # Shape after: (batch_size, seq_len, kv_head_num * n_repeat, head_dim)
            # where (kv_head_num * n_repeat) == ulysses_size
            key = torch.repeat_interleave(key, dim=2, repeats=n_repeat)
            value = torch.repeat_interleave(value, dim=2, repeats=n_repeat)

        if query.ndim == 4 and query.size(0) == 1:
            query, key, value = query.squeeze(0), key.squeeze(0), value.squeeze(0)
            query = gather_seq_scatter_heads(
                query, seq_dim=0, head_dim=1, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            key = gather_seq_scatter_heads(
                key, seq_dim=0, head_dim=1, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            value = gather_seq_scatter_heads(
                value, seq_dim=0, head_dim=1, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            query, key, value = query.unsqueeze(0), key.unsqueeze(0), value.unsqueeze(0)
        else:
            query = gather_seq_scatter_heads(
                query, seq_dim=1, head_dim=2, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            key = gather_seq_scatter_heads(
                key, seq_dim=1, head_dim=2, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )
            value = gather_seq_scatter_heads(
                value, seq_dim=1, head_dim=2, group=ulysses_group, unpadded_dim_size=unpadded_seq_len
            )

    # Only after all_to_all we got the full seq_len
    seq_len = query.shape[1]

    if is_seed_kernels_available() and implementation is not None:
        attn_output: torch.Tensor = seed_flash_attention_forward(
            query,
            key,
            value,
            attention_mask,
            query_length=seq_len,
            is_causal=module.is_causal,
            dropout=dropout,
            position_ids=position_ids,
            softmax_scale=scaling,
            sliding_window=sliding_window,
            softcap=softcap,
            use_top_left_mask=False,
            implementation=implementation,
            cu_seqlens=kwargs.get("cu_seq_lens_q", None),
            max_seqlen=kwargs.get("max_length_q", None),
            **kwargs,
        )
    else:
        assert implementation is None, (
            f"You set {implementation=} but seed_kernels is not installed. Check --model.attn_implementation."
        )
        attn_output: torch.Tensor = _flash_attention_forward(
            query,
            key,
            value,
            attention_mask,
            query_length=seq_len,
            is_causal=module.is_causal,
            dropout=dropout,
            position_ids=position_ids,
            softmax_scale=scaling,
            sliding_window=sliding_window,
            softcap=softcap,
            use_top_left_mask=False,
            implementation="flash_attention_2",
            **kwargs,
        )

    # Ulysses patch
    if ulysses_enabled and not skip_ulysses:
        ulysses_group = get_parallel_state().ulysses_group
        if attn_output.ndim == 4 and attn_output.size(0) == 1:
            attn_output = attn_output.squeeze(0)
            attn_output = gather_heads_scatter_seq(attn_output, seq_dim=0, head_dim=1, group=ulysses_group)
            attn_output = attn_output.unsqueeze(0)
        else:
            attn_output = gather_heads_scatter_seq(attn_output, seq_dim=1, head_dim=2, group=ulysses_group)

    return attn_output, None



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ...loader import MODEL_CONFIG_REGISTRY, MODELING_REGISTRY


@MODEL_CONFIG_REGISTRY.register("deepseek_v3")
def register_deepseek_v3_config():
    from .configuration_deepseek import DeepseekV3Config

    return DeepseekV3Config


@MODELING_REGISTRY.register("deepseek_v3")
def register_deepseek_v3_modeling(architecture: str):
    from .modeling_deepseek import DeepseekV3ForCausalLM, DeepseekV3ForSequenceClassification, DeepseekV3Model

    if "ForCausalLM" in architecture:
        return DeepseekV3ForCausalLM
    elif "ForSequenceClassification" in architecture:
        return DeepseekV3ForSequenceClassification
    elif "Model" in architecture:
        return DeepseekV3Model
    else:  # None
        return DeepseekV3ForCausalLM



import json
import os
import time
from collections import defaultdict
from dataclasses import asdict, dataclass, field
from functools import partial
from typing import Any, Dict, List, Optional

import torch
import wandb
from torch import distributed as dist
from tqdm import trange

from veomni.checkpoint import build_checkpointer, ckpt_to_state_dict
from veomni.data import (
    OmniDataCollatorWithPacking,
    OmniDataCollatorWithPadding,
    OmniSequenceShardCollator,
    build_dataloader,
    build_dataset,
    build_multimodal_chat_template,
)
from veomni.data.constants import IGNORE_INDEX
from veomni.data.multimodal.multimodal_transform import encode_multimodal_sample
from veomni.distributed.clip_grad_norm import veomni_clip_grad_norm
from veomni.distributed.offloading import build_activation_offloading_context
from veomni.distributed.parallel_state import get_parallel_state, init_parallel_state
from veomni.distributed.torch_parallelize import build_parallelize_model
from veomni.models import save_model_assets, save_model_weights
from veomni.models.seed_omni import SeedOmniModel, build_omni_model, build_omni_processor
from veomni.optim import build_lr_scheduler, build_optimizer
from veomni.utils import helper
from veomni.utils.arguments import DataArguments, ModelArguments, TrainingArguments, parse_args
from veomni.utils.device import get_device_type, get_torch_device, synchronize
from veomni.utils.dist_utils import all_reduce
from veomni.utils.model_utils import pretty_print_trainable_parameters


logger = helper.create_logger(__name__)

MAX_PIXELS = 768 * 28 * 28


@dataclass
class MyDataArguments(DataArguments):
    max_image_nums: Optional[int] = field(
        default=None,
        metadata={"help": "The max number of images in the sample."},
    )
    max_pixels: int = field(
        default=MAX_PIXELS,
        metadata={"help": "The max pixel numbers of the image."},
    )
    max_pixel_size: int = field(
        default=None,
        metadata={"help": "The max pixel size (height/width) of the image."},
    )
    scale_factor: int = field(
        default=None,
        metadata={"help": "Scale factor of the input image."},
    )
    generation_ratio: float = field(
        default=0.0,
        metadata={"help": "Ratio for instruction tuning data change to generation data."},
    )


@dataclass
class MyTrainingArguments(TrainingArguments):
    freeze_encoder: bool = field(
        default=False,
        metadata={"help": "Whether or not to freeze the encoder parameters."},
    )
    freeze_encoder_all: bool = field(
        default=False,
        metadata={"help": "Whether or not to freeze the whole encoder parameters."},
    )
    freeze_decoder: bool = field(
        default=False,
        metadata={"help": "Whether or not to freeze the decoder parameters."},
    )
    freeze_foundation: bool = field(
        default=False,
        metadata={"help": "Whether or not to freeze the foundation parameters."},
    )
    save_initial_model: bool = field(
        default=False,
        metadata={"help": "Whether or not to save the initial model."},
    )


@dataclass
class Arguments:
    model: "ModelArguments" = field(default_factory=ModelArguments)
    data: "MyDataArguments" = field(default_factory=MyDataArguments)
    train: "MyTrainingArguments" = field(default_factory=MyTrainingArguments)


def main():
    args = parse_args(Arguments)
    logger.info(f"Process rank: {args.train.global_rank}, world size: {args.train.world_size}")
    logger.info_rank0(json.dumps(asdict(args), indent=2))
    get_torch_device().set_device(f"{get_device_type()}:{args.train.local_rank}")
    dist.init_process_group()
    helper.set_seed(args.train.seed, args.train.enable_full_determinism)
    if args.train.local_rank == 0:
        helper.enable_third_party_logging()

    Checkpointer = build_checkpointer(dist_backend=args.train.data_parallel_mode, ckpt_manager=args.train.ckpt_manager)

    init_parallel_state(
        dp_size=args.train.data_parallel_size,
        dp_replicate_size=args.train.data_parallel_replicate_size,
        dp_shard_size=args.train.data_parallel_shard_size,
        tp_size=args.train.tensor_parallel_size,
        ep_size=args.train.expert_parallel_size,
        pp_size=args.train.pipeline_parallel_size,
        cp_size=args.train.context_parallel_size,
        ulysses_size=args.train.ulysses_parallel_size,
        dp_mode=args.train.data_parallel_mode,
    )
    logger.info_rank0("Prepare model")
    model: SeedOmniModel = build_omni_model(
        config_path=args.model.config_path,
        weights_path=args.model.model_path,
        torch_dtype="float32" if args.train.enable_mixed_precision else "bfloat16",
        encoders=args.model.encoders,
        decoders=args.model.decoders,
        input_encoder=args.model.input_encoder,
        output_encoder=args.model.output_encoder,
        init_device=args.train.init_device,
    )

    logger.info_rank0("Prepare data")
    processor = build_omni_processor(
        config_path=args.model.config_path,
        tokenizer_path=args.model.tokenizer_path,
        encoders=args.model.encoders,
        decoders=args.model.decoders,
        input_encoder=args.model.input_encoder,
        output_encoder=args.model.output_encoder,
        encode_target=args.model.encode_target,
        max_pixels=args.data.max_pixels,
    )
    chat_template = build_multimodal_chat_template(args.data.chat_template, processor.tokenizer)
    position_id_func = model.get_position_id_func()
    modality_info = model.get_modality()
    transform = partial(
        encode_multimodal_sample,
        processor=processor,
        chat_template=chat_template,
        position_id_func=position_id_func,
        modality_info=modality_info,
        max_image_nums=args.data.max_image_nums,
        max_pixel_size=args.data.max_pixel_size,
        generation_ratio=args.data.generation_ratio,
        scale_factor=args.data.scale_factor,
    )

    if args.train.rmpad:
        raise ValueError("Qwen2-VL does not support rmpad. Use `rmpad_with_pos_ids` instead.")

    data_collate_fn = []

    # TODO: config by model
    if args.train.rmpad_with_pos_ids:
        data_collate_fn.append(
            OmniDataCollatorWithPacking(
                packing_features=[
                    "input_ids",
                    "attention_mask",
                    "labels",
                    "position_ids",
                    "image_input_mask",
                    "image_output_mask",
                ],
                concat_features=[
                    "image_input_features",
                    "image_input_grid_thw",
                    "image_output_features",
                    "image_output_grid_thw",
                ],
            )
        )
    else:
        data_collate_fn.append(
            OmniDataCollatorWithPadding(
                concat_features={
                    "image_input_features": 0,
                    "image_input_grid_thw": 0,
                    "image_output_features": 0,
                    "image_output_grid_thw": 0,
                },
                padding_features={
                    "input_ids": 0,
                    "attention_mask": 0,
                    "labels": IGNORE_INDEX,
                    "position_ids": 0,
                    "image_input_mask": False,
                    "image_output_mask": False,
                },
            )
        )
    if get_parallel_state().sp_enabled:
        data_collate_fn.append(
            OmniSequenceShardCollator(
                sp_slice_features={
                    "input_ids": -1,
                    "labels": -1,
                    "image_input_features": 0,
                    "image_output_features": 0,
                },
                padding_features={
                    "input_ids": 0,
                    "attention_mask": 0,
                    "labels": IGNORE_INDEX,
                    "position_ids": 0,
                    "image_input_features": 0,
                    "image_output_features": 0,
                    "image_input_mask": False,
                    "image_output_mask": False,
                },
                padding_scale={
                    "image_input_features": 4,
                    "image_output_features": 1,
                },
                rmpad_with_pos_ids=args.train.rmpad_with_pos_ids,
            )
        )

    train_dataset = build_dataset(
        dataset_name=args.data.dataset_name,
        transform=transform,
        dataloader_batch_size=args.train.dataloader_batch_size,
        seed=args.train.seed,
        **asdict(args.data),
    )
    dataset_length = None if not hasattr(train_dataset, "__len__") else len(train_dataset)
    if args.data.datasets_type == "mapping":
        dataset_length = dataset_length / args.train.data_parallel_size
    args.train.compute_train_steps(args.data.max_seq_len, args.data.train_size, dataset_length)

    train_dataloader = build_dataloader(
        dataloader_type=args.data.dataloader_type,
        dataset=train_dataset,
        micro_batch_size=args.train.micro_batch_size,
        global_batch_size=args.train.global_batch_size,
        dataloader_batch_size=args.train.dataloader_batch_size,
        seed=args.train.seed,
        collate_fn=data_collate_fn,
        max_seq_len=args.data.max_seq_len,
        train_steps=args.train.train_steps,
        rmpad=args.train.rmpad,
        rmpad_with_pos_ids=args.train.rmpad_with_pos_ids,
        bsz_warmup_ratio=args.train.bsz_warmup_ratio,
        dyn_bsz_margin=args.train.dyn_bsz_margin,
        dyn_bsz_buffer_size=args.train.dyn_bsz_buffer_size,
        num_workers=args.data.num_workers,
        drop_last=args.data.drop_last,
        pin_memory=args.data.pin_memory,
        prefetch_factor=args.data.prefetch_factor,
    )

    freeze_any = False
    if args.train.freeze_encoder:
        if args.train.freeze_encoder_all:
            model.encoder.requires_grad_(False)
        else:
            model.encoder.set_projector_trainable_only()
        freeze_any = True
    if args.train.freeze_decoder:
        model.decoder.set_projector_trainable_only()
        freeze_any = True
    if args.train.freeze_foundation:
        model.foundation.requires_grad_(False)
        freeze_any = True
    pretty_print_trainable_parameters(model)

    fsdp_kwargs = {}
    if freeze_any:
        if args.train.data_parallel_mode == "fsdp1":
            fsdp_kwargs["use_orig_params"] = True

    model_config = model.config
    helper.print_device_mem_info("VRAM usage after building model")

    if args.train.save_initial_model:
        if args.train.global_rank == 0:
            save_model_weights(args.train.output_dir, model.state_dict(), model_assets=[model_config, processor])

        dist.barrier()
        return

    model = build_parallelize_model(
        model,
        enable_full_shard=args.train.enable_full_shard,
        enable_mixed_precision=args.train.enable_mixed_precision,
        enable_gradient_checkpointing=args.train.enable_gradient_checkpointing,
        init_device=args.train.init_device,
        enable_fsdp_offload=args.train.enable_fsdp_offload,
        fsdp_kwargs=fsdp_kwargs,
        basic_modules=model._no_split_modules,
        enable_reentrant=args.train.enable_reentrant,
        enable_forward_prefetch=args.train.enable_forward_prefetch,
    )
    optimizer = build_optimizer(
        model,
        lr=args.train.lr,
        weight_decay=args.train.weight_decay,
        fused=False,
        optimizer_type=args.train.optimizer,
    )
    lr_scheduler = build_lr_scheduler(
        optimizer,
        train_steps=args.train.train_steps * args.train.num_train_epochs,
        lr=args.train.lr,
        lr_min=args.train.lr_min,
        lr_decay_style=args.train.lr_decay_style,
        lr_decay_ratio=args.train.lr_decay_ratio,
        lr_warmup_ratio=args.train.lr_warmup_ratio,
        lr_start=args.train.lr_start,
    )

    if args.train.global_rank == 0:
        if args.train.use_wandb:
            wandb.init(
                project=args.train.wandb_project,
                name=args.train.wandb_name,
                config={**vars(args.model), **vars(args.data), **vars(args.train)},  # flatten dict
            )

        model_assets = [model_config, processor]
        save_model_assets(args.train.model_assets_dir, model_assets)

    if args.train.profile_this_rank:
        profiler = helper.create_profiler(
            start_step=args.train.profile_start_step,
            end_step=args.train.profile_end_step,
            trace_dir=args.train.profile_trace_dir,
            record_shapes=args.train.profile_record_shapes,
            profile_memory=args.train.profile_profile_memory,
            with_stack=args.train.profile_with_stack,
            global_rank=args.train.global_rank,
        )
        profiler.start()

    start_epoch, start_step, global_step = 0, 0, 0
    save_checkpoint_path = None
    environ_meter = helper.EnvironMeter(
        config=model_config,
        global_batch_size=args.train.global_batch_size,
        rmpad=args.train.rmpad,
        rmpad_with_pos_ids=args.train.rmpad_with_pos_ids,
        empty_cache_steps=args.train.empty_cache_steps,
        enable_multisource=args.data.enable_multisource,
        dataloader=train_dataloader,
        data_path=args.data.train_path,
    )

    if args.train.load_checkpoint_path:
        state = {"model": model, "optimizer": optimizer, "extra_state": {}}  # cannot be None
        Checkpointer.load(args.train.load_checkpoint_path, state)
        global_step = state["extra_state"]["global_step"]
        start_epoch = global_step // args.train.train_steps
        start_step = global_step % args.train.train_steps
        lr_scheduler_state = state["extra_state"]["lr_scheduler"]
        lr_scheduler.load_state_dict(lr_scheduler_state)
        train_dataloader.load_state_dict(state["extra_state"]["train_dataloader"])
        environ_meter.load_state_dict(state["extra_state"]["environ_meter"])
        torch.set_rng_state(state["extra_state"]["torch_rng_state"])
        del state
        if start_step == 0:  # resume at the end of epoch
            iter(train_dataloader)  # clear resume state and prefetch data

        dist.barrier()
        logger.info_rank0(f"Load distributed checkpoint from {args.train.load_checkpoint_path} successfully!")

    helper.empty_cache()
    model_fwd_context, model_bwd_context = build_activation_offloading_context(
        args.train.enable_activation_offload, args.train.enable_gradient_checkpointing, args.train.activation_gpu_limit
    )
    model.train()
    logger.info(
        f"rank{args.train.local_rank} Start training, train_steps: {args.train.train_steps}, epochs: {args.train.num_train_epochs}"
    )
    for epoch in range(start_epoch, args.train.num_train_epochs):
        if hasattr(train_dataloader, "set_epoch"):
            train_dataloader.set_epoch(epoch)

        data_loader_tqdm = trange(
            args.train.train_steps,
            desc=f"Epoch {epoch + 1}/{args.train.num_train_epochs}",
            total=args.train.train_steps,
            initial=start_step,
            disable=args.train.local_rank != 0,
        )
        data_iterator = iter(train_dataloader)
        for _ in range(start_step, args.train.train_steps):
            global_step += 1

            try:
                micro_batches: List[Dict[str, Any]] = next(data_iterator)
            except StopIteration:
                logger.info(f"epoch:{epoch} Dataloader finished with drop_last {args.data.drop_last}")
                break

            if global_step == 1:
                helper.print_example(example=micro_batches[0], rank=args.train.local_rank)

            total_loss = 0
            total_losses = defaultdict(int)
            synchronize()
            start_time = time.time()
            for micro_batch in micro_batches:
                environ_meter.add(micro_batch)
                if args.data.enable_multisource:
                    micro_batch.pop("ds_idx", None)
                    micro_batch.pop("cur_token_num", None)
                    micro_batch.pop("source_name", None)

                micro_batch = {
                    k: v.to(get_device_type(), non_blocking=True) if isinstance(v, torch.Tensor) else v
                    for k, v in micro_batch.items()
                }
                with model_fwd_context:
                    model_outputs = model(**micro_batch, use_cache=False)

                loss: "torch.Tensor" = model_outputs.loss / len(micro_batches)

                with model_bwd_context:
                    loss.backward()

                total_loss += loss.item()
                losses = model_outputs.losses
                for key, v in losses.items():
                    total_losses[key] += v / len(micro_batches)

                del micro_batch

            grad_norm = veomni_clip_grad_norm(model, args.train.max_grad_norm)

            optimizer.step()
            lr_scheduler.step()
            optimizer.zero_grad()
            if hasattr(grad_norm, "full_tensor"):
                grad_norm = grad_norm.full_tensor().item()

            # collect loss across data parallel group
            total_loss, grad_norm = all_reduce((total_loss, grad_norm), group=get_parallel_state().fsdp_group)
            for key, v in total_losses.items():
                total_losses[key] = all_reduce((v), group=get_parallel_state().fsdp_group)
            synchronize()
            delta_time = time.time() - start_time
            lr = max(lr_scheduler.get_last_lr())
            train_metrics = environ_meter.step(delta_time=delta_time, global_step=global_step)

            step_info = total_losses
            step_info.update(
                {
                    "loss": total_loss,
                    "grad_norm": grad_norm,
                    "lr": lr,
                }
            )
            data_loader_tqdm.set_postfix_str({k: f"{v:.2f}" for k, v in step_info.items() if k != "lr"})
            data_loader_tqdm.update()

            if args.train.global_rank == 0:
                if args.train.use_wandb:
                    train_metrics.update({f"training/{k}": v for k, v in step_info.items()})
                    wandb.log(train_metrics, step=global_step)

            if args.train.profile_this_rank and global_step <= args.train.profile_end_step:
                profiler.step()
                if global_step == args.train.profile_end_step:
                    profiler.stop()
                    helper.upload_trace(args.train.wandb_project, args.train.wandb_name, args.train.profile_trace_dir)

            if args.train.save_steps and global_step % args.train.save_steps == 0:
                helper.empty_cache()
                save_checkpoint_path = os.path.join(args.train.save_checkpoint_path, f"global_step_{global_step}")
                state = {
                    "model": model,
                    "optimizer": optimizer,
                    "extra_state": {
                        "global_step": global_step,
                        "lr_scheduler": lr_scheduler.state_dict(),
                        "train_dataloader": train_dataloader.state_dict(),
                        "environ_meter": environ_meter.state_dict(),
                        "torch_rng_state": torch.get_rng_state(),
                    },
                }
                Checkpointer.save(args.train.save_checkpoint_path, state, global_steps=global_step)
                dist.barrier()
                logger.info_rank0(f"Distributed checkpoint saved at {save_checkpoint_path} successfully!")

        data_loader_tqdm.close()
        start_step = 0
        helper.print_device_mem_info(f"VRAM usage after epoch {epoch + 1}")
        if args.train.save_epochs and (epoch + 1) % args.train.save_epochs == 0:
            helper.empty_cache()
            save_checkpoint_path = os.path.join(args.train.save_checkpoint_path, f"global_step_{global_step}")
            state = {
                "model": model,
                "optimizer": optimizer,
                "extra_state": {
                    "global_step": global_step,
                    "lr_scheduler": lr_scheduler.state_dict(),
                    "train_dataloader": train_dataloader.state_dict(),
                    "environ_meter": environ_meter.state_dict(),
                    "torch_rng_state": torch.get_rng_state(),
                },
            }
            Checkpointer.save(args.train.save_checkpoint_path, state, global_steps=global_step)
            dist.barrier()
            logger.info_rank0(f"Distributed checkpoint saved at {save_checkpoint_path} successfully!")

    synchronize()
    # release memory
    del optimizer, lr_scheduler
    helper.empty_cache()
    # save model in huggingface's format
    if args.train.global_rank == 0 and args.train.save_hf_weights and save_checkpoint_path is not None:
        hf_weights_path = os.path.join(save_checkpoint_path, "hf_ckpt")
        model_state_dict = ckpt_to_state_dict(
            save_checkpoint_path=save_checkpoint_path,
            output_dir=args.train.output_dir,
            ckpt_manager=args.train.ckpt_manager,
        )
        save_model_weights(hf_weights_path, model_state_dict, model_assets=model_assets)
        logger.info_rank0(f"Huggingface checkpoint saved at {args.train.output_dir} successfully!")

    dist.barrier()
    dist.destroy_process_group()


if __name__ == "__main__":
    main()
